# encoding: utf-8

"""
Composers for default chart XML for various chart types.
"""

from __future__ import absolute_import, print_function, unicode_literals

from copy import deepcopy
from xml.sax.saxutils import escape

from ..compat import to_unicode
from ..enum.chart import XL_CHART_TYPE
from ..oxml import parse_xml
from ..oxml.ns import nsdecls


def ChartXmlWriter(chart_type, chart_data):
    """
    Factory function returning appropriate XML writer object for
    *chart_type*, loaded with *chart_type* and *chart_data*.
    """
    XL_CT = XL_CHART_TYPE
    try:
        BuilderCls = {
            XL_CT.AREA: _AreaChartXmlWriter,
            XL_CT.AREA_STACKED: _AreaChartXmlWriter,
            XL_CT.AREA_STACKED_100: _AreaChartXmlWriter,
            XL_CT.BAR_CLUSTERED: _BarChartXmlWriter,
            XL_CT.BAR_STACKED: _BarChartXmlWriter,
            XL_CT.BAR_STACKED_100: _BarChartXmlWriter,
            XL_CT.BUBBLE: _BubbleChartXmlWriter,
            XL_CT.BUBBLE_THREE_D_EFFECT: _BubbleChartXmlWriter,
            XL_CT.COLUMN_CLUSTERED: _BarChartXmlWriter,
            XL_CT.COLUMN_STACKED: _BarChartXmlWriter,
            XL_CT.COLUMN_STACKED_100: _BarChartXmlWriter,
            XL_CT.DOUGHNUT: _DoughnutChartXmlWriter,
            XL_CT.DOUGHNUT_EXPLODED: _DoughnutChartXmlWriter,
            XL_CT.LINE: _LineChartXmlWriter,
            XL_CT.LINE_MARKERS: _LineChartXmlWriter,
            XL_CT.LINE_MARKERS_STACKED: _LineChartXmlWriter,
            XL_CT.LINE_MARKERS_STACKED_100: _LineChartXmlWriter,
            XL_CT.LINE_STACKED: _LineChartXmlWriter,
            XL_CT.LINE_STACKED_100: _LineChartXmlWriter,
            XL_CT.PIE: _PieChartXmlWriter,
            XL_CT.PIE_EXPLODED: _PieChartXmlWriter,
            XL_CT.RADAR: _RadarChartXmlWriter,
            XL_CT.RADAR_FILLED: _RadarChartXmlWriter,
            XL_CT.RADAR_MARKERS: _RadarChartXmlWriter,
            XL_CT.XY_SCATTER: _XyChartXmlWriter,
            XL_CT.XY_SCATTER_LINES: _XyChartXmlWriter,
            XL_CT.XY_SCATTER_LINES_NO_MARKERS: _XyChartXmlWriter,
            XL_CT.XY_SCATTER_SMOOTH: _XyChartXmlWriter,
            XL_CT.XY_SCATTER_SMOOTH_NO_MARKERS: _XyChartXmlWriter,
        }[chart_type]
    except KeyError:
        raise NotImplementedError(
            "XML writer for chart type %s not yet implemented" % chart_type
        )
    return BuilderCls(chart_type, chart_data)


def SeriesXmlRewriterFactory(chart_type, chart_data):
    """
    Return a |_BaseSeriesXmlRewriter| subclass appropriate to *chart_type*.
    """
    XL_CT = XL_CHART_TYPE

    RewriterCls = {
        # There are 73 distinct chart types, only specify non-category
        # types, others default to _CategorySeriesXmlRewriter. Stock-type
        # charts are multi-plot charts, so no guaratees on how they turn
        # out.
        XL_CT.BUBBLE: _BubbleSeriesXmlRewriter,
        XL_CT.BUBBLE_THREE_D_EFFECT: _BubbleSeriesXmlRewriter,
        XL_CT.XY_SCATTER: _XySeriesXmlRewriter,
        XL_CT.XY_SCATTER_LINES: _XySeriesXmlRewriter,
        XL_CT.XY_SCATTER_LINES_NO_MARKERS: _XySeriesXmlRewriter,
        XL_CT.XY_SCATTER_SMOOTH: _XySeriesXmlRewriter,
        XL_CT.XY_SCATTER_SMOOTH_NO_MARKERS: _XySeriesXmlRewriter,
    }.get(chart_type, _CategorySeriesXmlRewriter)

    return RewriterCls(chart_data)


class _BaseChartXmlWriter(object):
    """
    Generates XML text (unicode) for a default chart, like the one added by
    PowerPoint when you click the *Add Column Chart* button on the ribbon.
    Differentiated XML for different chart types is provided by subclasses.
    """

    def __init__(self, chart_type, series_seq):
        super(_BaseChartXmlWriter, self).__init__()
        self._chart_type = chart_type
        self._chart_data = series_seq
        self._series_seq = list(series_seq)

    @property
    def xml(self):
        """
        The full XML stream for the chart specified by this chart builder, as
        unicode text. This method must be overridden by each subclass.
        """
        raise NotImplementedError("must be implemented by all subclasses")


class _BaseSeriesXmlWriter(object):
    """
    Provides shared members for series XML writers.
    """

    def __init__(self, series, date_1904=False):
        super(_BaseSeriesXmlWriter, self).__init__()
        self._series = series
        self._date_1904 = date_1904

    @property
    def name(self):
        """
        The XML-escaped name for this series.
        """
        return escape(self._series.name)

    def numRef_xml(self, wksht_ref, number_format, values):
        """
        Return the ``<c:numRef>`` element specified by the parameters as
        unicode text.
        """
        pt_xml = self.pt_xml(values)
        return (
            "            <c:numRef>\n"
            "              <c:f>{wksht_ref}</c:f>\n"
            "              <c:numCache>\n"
            "                <c:formatCode>{number_format}</c:formatCode>\n"
            "{pt_xml}"
            "              </c:numCache>\n"
            "            </c:numRef>\n"
        ).format(
            **{"wksht_ref": wksht_ref, "number_format": number_format, "pt_xml": pt_xml}
        )

    def pt_xml(self, values):
        """
        Return the ``<c:ptCount>`` and sequence of ``<c:pt>`` elements
        corresponding to *values* as a single unicode text string.
        `c:ptCount` refers to the number of `c:pt` elements in this sequence.
        The `idx` attribute value for `c:pt` elements locates the data point
        in the overall data point sequence of the chart and is started at
        *offset*.
        """
        xml = ('                <c:ptCount val="{pt_count}"/>\n').format(
            pt_count=len(values)
        )

        pt_tmpl = (
            '                <c:pt idx="{idx}">\n'
            "                  <c:v>{value}</c:v>\n"
            "                </c:pt>\n"
        )
        for idx, value in enumerate(values):
            if value is None:
                continue
            xml += pt_tmpl.format(idx=idx, value=value)

        return xml

    @property
    def tx(self):
        """
        Return a ``<c:tx>`` oxml element for this series, containing the
        series name.
        """
        xml = self._tx_tmpl.format(
            **{
                "wksht_ref": self._series.name_ref,
                "series_name": self.name,
                "nsdecls": " %s" % nsdecls("c"),
            }
        )
        return parse_xml(xml)

    @property
    def tx_xml(self):
        """
        Return the ``<c:tx>`` (tx is short for 'text') element for this
        series as unicode text. This element contains the series name.
        """
        return self._tx_tmpl.format(
            **{
                "wksht_ref": self._series.name_ref,
                "series_name": self.name,
                "nsdecls": "",
            }
        )

    @property
    def _tx_tmpl(self):
        """
        The string formatting template for the ``<c:tx>`` element for this
        series, containing the series title and spreadsheet range reference.
        """
        return (
            "          <c:tx{nsdecls}>\n"
            "            <c:strRef>\n"
            "              <c:f>{wksht_ref}</c:f>\n"
            "              <c:strCache>\n"
            '                <c:ptCount val="1"/>\n'
            '                <c:pt idx="0">\n'
            "                  <c:v>{series_name}</c:v>\n"
            "                </c:pt>\n"
            "              </c:strCache>\n"
            "            </c:strRef>\n"
            "          </c:tx>\n"
        )


class _BaseSeriesXmlRewriter(object):
    """
    Base class for series XML rewriters.
    """

    def __init__(self, chart_data):
        super(_BaseSeriesXmlRewriter, self).__init__()
        self._chart_data = chart_data

    def replace_series_data(self, chartSpace):
        """
        Rewrite the series data under *chartSpace* using the chart data
        contents. All series-level formatting is left undisturbed. If
        the chart data contains fewer series than *chartSpace*, the extra
        series in *chartSpace* are deleted. If *chart_data* contains more
        series than the *chartSpace* element, new series are added to the
        last plot in the chart and series formatting is "cloned" from the
        last series in that plot.
        """
        plotArea, date_1904 = chartSpace.plotArea, chartSpace.date_1904
        chart_data = self._chart_data
        self._adjust_ser_count(plotArea, len(chart_data))
        for ser, series_data in zip(plotArea.sers, chart_data):
            self._rewrite_ser_data(ser, series_data, date_1904)

    def _add_cloned_sers(self, plotArea, count):
        """
        Add `c:ser` elements to the last xChart element in *plotArea*, cloned
        from the last `c:ser` child of that last xChart.
        """

        def clone_ser(ser):
            new_ser = deepcopy(ser)
            new_ser.idx.val = plotArea.next_idx
            new_ser.order.val = plotArea.next_order
            ser.addnext(new_ser)
            return new_ser

        last_ser = plotArea.last_ser
        for _ in range(count):
            last_ser = clone_ser(last_ser)

    def _adjust_ser_count(self, plotArea, new_ser_count):
        """
        Adjust the number of c:ser elements in *plotArea* to *new_ser_count*.
        Excess c:ser elements are deleted from the end, along with any xChart
        elements that are left empty as a result. Series elements are
        considered in xChart + series order. Any new c:ser elements required
        are added to the last xChart element and cloned from the last c:ser
        element in that xChart.
        """
        ser_count_diff = new_ser_count - len(plotArea.sers)
        if ser_count_diff > 0:
            self._add_cloned_sers(plotArea, ser_count_diff)
        elif ser_count_diff < 0:
            self._trim_ser_count_by(plotArea, abs(ser_count_diff))

    def _rewrite_ser_data(self, ser, series_data, date_1904):
        """
        Rewrite selected child elements of *ser* based on the values in
        *series_data*.
        """
        raise NotImplementedError("must be implemented by each subclass")

    def _trim_ser_count_by(self, plotArea, count):
        """
        Remove the last *count* ser elements from *plotArea*. Any xChart
        elements having no ser child elements after trimming are also
        removed.
        """
        extra_sers = plotArea.sers[-count:]
        for ser in extra_sers:
            parent = ser.getparent()
            parent.remove(ser)
        extra_xCharts = [
            xChart for xChart in plotArea.iter_xCharts() if len(xChart.sers) == 0
        ]
        for xChart in extra_xCharts:
            parent = xChart.getparent()
            parent.remove(xChart)


class _AreaChartXmlWriter(_BaseChartXmlWriter):
    """
    Provides specialized methods particular to the ``<c:areaChart>`` element.
    """

    @property
    def xml(self):
        return (
            "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
            '<c:chartSpace xmlns:c="http://schemas.openxmlformats.org/drawin'
            'gml/2006/chart" xmlns:a="http://schemas.openxmlformats.org/draw'
            'ingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/off'
            'iceDocument/2006/relationships">\n'
            '  <c:date1904 val="0"/>\n'
            '  <c:roundedCorners val="0"/>\n'
            "  <c:chart>\n"
            '    <c:autoTitleDeleted val="0"/>\n'
            "    <c:plotArea>\n"
            "      <c:layout/>\n"
            "      <c:areaChart>\n"
            "{grouping_xml}"
            '        <c:varyColors val="0"/>\n'
            "{ser_xml}"
            "        <c:dLbls>\n"
            '          <c:showLegendKey val="0"/>\n'
            '          <c:showVal val="0"/>\n'
            '          <c:showCatName val="0"/>\n'
            '          <c:showSerName val="0"/>\n'
            '          <c:showPercent val="0"/>\n'
            '          <c:showBubbleSize val="0"/>\n'
            "        </c:dLbls>\n"
            '        <c:axId val="-2101159928"/>\n'
            '        <c:axId val="-2100718248"/>\n'
            "      </c:areaChart>\n"
            "{cat_ax_xml}"
            "      <c:valAx>\n"
            '        <c:axId val="-2100718248"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="l"/>\n'
            "        <c:majorGridlines/>\n"
            '        <c:numFmt formatCode="General" sourceLinked="1"/>\n'
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="-2101159928"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:crossBetween val="midCat"/>\n'
            "      </c:valAx>\n"
            "    </c:plotArea>\n"
            "    <c:legend>\n"
            '      <c:legendPos val="r"/>\n'
            "      <c:layout/>\n"
            '      <c:overlay val="0"/>\n'
            "    </c:legend>\n"
            '    <c:plotVisOnly val="1"/>\n'
            '    <c:dispBlanksAs val="zero"/>\n'
            '    <c:showDLblsOverMax val="0"/>\n'
            "  </c:chart>\n"
            "  <c:txPr>\n"
            "    <a:bodyPr/>\n"
            "    <a:lstStyle/>\n"
            "    <a:p>\n"
            "      <a:pPr>\n"
            '        <a:defRPr sz="1800"/>\n'
            "      </a:pPr>\n"
            "      <a:endParaRPr/>\n"
            "    </a:p>\n"
            "  </c:txPr>\n"
            "</c:chartSpace>\n"
        ).format(
            **{
                "grouping_xml": self._grouping_xml,
                "ser_xml": self._ser_xml,
                "cat_ax_xml": self._cat_ax_xml,
            }
        )

    @property
    def _cat_ax_xml(self):
        categories = self._chart_data.categories

        if categories.are_dates:
            return (
                "      <c:dateAx>\n"
                '        <c:axId val="-2101159928"/>\n'
                "        <c:scaling>\n"
                '          <c:orientation val="minMax"/>\n'
                "        </c:scaling>\n"
                '        <c:delete val="0"/>\n'
                '        <c:axPos val="b"/>\n'
                '        <c:numFmt formatCode="{nf}" sourceLinked="1"/>\n'
                '        <c:majorTickMark val="out"/>\n'
                '        <c:minorTickMark val="none"/>\n'
                '        <c:tickLblPos val="nextTo"/>\n'
                '        <c:crossAx val="-2100718248"/>\n'
                '        <c:crosses val="autoZero"/>\n'
                '        <c:auto val="1"/>\n'
                '        <c:lblOffset val="100"/>\n'
                '        <c:baseTimeUnit val="days"/>\n'
                "      </c:dateAx>\n"
            ).format(**{"nf": categories.number_format})

        return (
            "      <c:catAx>\n"
            '        <c:axId val="-2101159928"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="b"/>\n'
            '        <c:numFmt formatCode="General" sourceLinked="1"/>\n'
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="-2100718248"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:auto val="1"/>\n'
            '        <c:lblAlgn val="ctr"/>\n'
            '        <c:lblOffset val="100"/>\n'
            '        <c:noMultiLvlLbl val="0"/>\n'
            "      </c:catAx>\n"
        )

    @property
    def _grouping_xml(self):
        val = {
            XL_CHART_TYPE.AREA: "standard",
            XL_CHART_TYPE.AREA_STACKED: "stacked",
            XL_CHART_TYPE.AREA_STACKED_100: "percentStacked",
        }[self._chart_type]
        return '        <c:grouping val="%s"/>\n' % val

    @property
    def _ser_xml(self):
        xml = ""
        for series in self._chart_data:
            xml_writer = _CategorySeriesXmlWriter(series)
            xml += (
                "        <c:ser>\n"
                '          <c:idx val="{ser_idx}"/>\n'
                '          <c:order val="{ser_order}"/>\n'
                "{tx_xml}"
                "{cat_xml}"
                "{val_xml}"
                "        </c:ser>\n"
            ).format(
                **{
                    "ser_idx": series.index,
                    "ser_order": series.index,
                    "tx_xml": xml_writer.tx_xml,
                    "cat_xml": xml_writer.cat_xml,
                    "val_xml": xml_writer.val_xml,
                }
            )
        return xml


class _BarChartXmlWriter(_BaseChartXmlWriter):
    """
    Provides specialized methods particular to the ``<c:barChart>`` element.
    """

    @property
    def xml(self):
        return (
            "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
            '<c:chartSpace xmlns:c="http://schemas.openxmlformats.org/drawin'
            'gml/2006/chart" xmlns:a="http://schemas.openxmlformats.org/draw'
            'ingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/off'
            'iceDocument/2006/relationships">\n'
            '  <c:date1904 val="0"/>\n'
            "  <c:chart>\n"
            '    <c:autoTitleDeleted val="0"/>\n'
            "    <c:plotArea>\n"
            "      <c:barChart>\n"
            "{barDir_xml}"
            "{grouping_xml}"
            "{ser_xml}"
            "{overlap_xml}"
            '        <c:axId val="-2068027336"/>\n'
            '        <c:axId val="-2113994440"/>\n'
            "      </c:barChart>\n"
            "{cat_ax_xml}"
            "      <c:valAx>\n"
            '        <c:axId val="-2113994440"/>\n'
            "        <c:scaling/>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="{val_ax_pos}"/>\n'
            "        <c:majorGridlines/>\n"
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="-2068027336"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            "      </c:valAx>\n"
            "    </c:plotArea>\n"
            '    <c:dispBlanksAs val="gap"/>\n'
            "  </c:chart>\n"
            "  <c:txPr>\n"
            "    <a:bodyPr/>\n"
            "    <a:lstStyle/>\n"
            "    <a:p>\n"
            "      <a:pPr>\n"
            '        <a:defRPr sz="1800"/>\n'
            "      </a:pPr>\n"
            '      <a:endParaRPr lang="en-US"/>\n'
            "    </a:p>\n"
            "  </c:txPr>\n"
            "</c:chartSpace>\n"
        ).format(
            **{
                "barDir_xml": self._barDir_xml,
                "grouping_xml": self._grouping_xml,
                "ser_xml": self._ser_xml,
                "overlap_xml": self._overlap_xml,
                "cat_ax_xml": self._cat_ax_xml,
                "val_ax_pos": self._val_ax_pos,
            }
        )

    @property
    def _barDir_xml(self):
        XL = XL_CHART_TYPE
        bar_types = (XL.BAR_CLUSTERED, XL.BAR_STACKED, XL.BAR_STACKED_100)
        col_types = (XL.COLUMN_CLUSTERED, XL.COLUMN_STACKED, XL.COLUMN_STACKED_100)
        if self._chart_type in bar_types:
            return '        <c:barDir val="bar"/>\n'
        elif self._chart_type in col_types:
            return '        <c:barDir val="col"/>\n'
        raise NotImplementedError(
            "no _barDir_xml() for chart type %s" % self._chart_type
        )

    @property
    def _cat_ax_pos(self):
        return {
            XL_CHART_TYPE.BAR_CLUSTERED: "l",
            XL_CHART_TYPE.BAR_STACKED: "l",
            XL_CHART_TYPE.BAR_STACKED_100: "l",
            XL_CHART_TYPE.COLUMN_CLUSTERED: "b",
            XL_CHART_TYPE.COLUMN_STACKED: "b",
            XL_CHART_TYPE.COLUMN_STACKED_100: "b",
        }[self._chart_type]

    @property
    def _cat_ax_xml(self):
        categories = self._chart_data.categories

        if categories.are_dates:
            return (
                "      <c:dateAx>\n"
                '        <c:axId val="-2068027336"/>\n'
                "        <c:scaling>\n"
                '          <c:orientation val="minMax"/>\n'
                "        </c:scaling>\n"
                '        <c:delete val="0"/>\n'
                '        <c:axPos val="{cat_ax_pos}"/>\n'
                '        <c:numFmt formatCode="{nf}" sourceLinked="1"/>\n'
                '        <c:majorTickMark val="out"/>\n'
                '        <c:minorTickMark val="none"/>\n'
                '        <c:tickLblPos val="nextTo"/>\n'
                '        <c:crossAx val="-2113994440"/>\n'
                '        <c:crosses val="autoZero"/>\n'
                '        <c:auto val="1"/>\n'
                '        <c:lblOffset val="100"/>\n'
                '        <c:baseTimeUnit val="days"/>\n'
                "      </c:dateAx>\n"
            ).format(**{"cat_ax_pos": self._cat_ax_pos, "nf": categories.number_format})

        return (
            "      <c:catAx>\n"
            '        <c:axId val="-2068027336"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="{cat_ax_pos}"/>\n'
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="-2113994440"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:auto val="1"/>\n'
            '        <c:lblAlgn val="ctr"/>\n'
            '        <c:lblOffset val="100"/>\n'
            '        <c:noMultiLvlLbl val="0"/>\n'
            "      </c:catAx>\n"
        ).format(**{"cat_ax_pos": self._cat_ax_pos})

    @property
    def _grouping_xml(self):
        XL = XL_CHART_TYPE
        clustered_types = (XL.BAR_CLUSTERED, XL.COLUMN_CLUSTERED)
        stacked_types = (XL.BAR_STACKED, XL.COLUMN_STACKED)
        percentStacked_types = (XL.BAR_STACKED_100, XL.COLUMN_STACKED_100)
        if self._chart_type in clustered_types:
            return '        <c:grouping val="clustered"/>\n'
        elif self._chart_type in stacked_types:
            return '        <c:grouping val="stacked"/>\n'
        elif self._chart_type in percentStacked_types:
            return '        <c:grouping val="percentStacked"/>\n'
        raise NotImplementedError(
            "no _grouping_xml() for chart type %s" % self._chart_type
        )

    @property
    def _overlap_xml(self):
        XL = XL_CHART_TYPE
        percentStacked_types = (
            XL.BAR_STACKED,
            XL.BAR_STACKED_100,
            XL.COLUMN_STACKED,
            XL.COLUMN_STACKED_100,
        )
        if self._chart_type in percentStacked_types:
            return '        <c:overlap val="100"/>\n'
        return ""

    @property
    def _ser_xml(self):
        xml = ""
        for series in self._chart_data:
            xml_writer = _CategorySeriesXmlWriter(series)
            xml += (
                "        <c:ser>\n"
                '          <c:idx val="{ser_idx}"/>\n'
                '          <c:order val="{ser_order}"/>\n'
                "{tx_xml}"
                "{cat_xml}"
                "{val_xml}"
                "        </c:ser>\n"
            ).format(
                **{
                    "ser_idx": series.index,
                    "ser_order": series.index,
                    "tx_xml": xml_writer.tx_xml,
                    "cat_xml": xml_writer.cat_xml,
                    "val_xml": xml_writer.val_xml,
                }
            )
        return xml

    @property
    def _val_ax_pos(self):
        return {
            XL_CHART_TYPE.BAR_CLUSTERED: "b",
            XL_CHART_TYPE.BAR_STACKED: "b",
            XL_CHART_TYPE.BAR_STACKED_100: "b",
            XL_CHART_TYPE.COLUMN_CLUSTERED: "l",
            XL_CHART_TYPE.COLUMN_STACKED: "l",
            XL_CHART_TYPE.COLUMN_STACKED_100: "l",
        }[self._chart_type]


class _DoughnutChartXmlWriter(_BaseChartXmlWriter):
    """
    Provides specialized methods particular to the ``<c:doughnutChart>``
    element.
    """

    @property
    def xml(self):
        return (
            "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
            '<c:chartSpace xmlns:c="http://schemas.openxmlformats.org/drawin'
            'gml/2006/chart" xmlns:a="http://schemas.openxmlformats.org/draw'
            'ingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/off'
            'iceDocument/2006/relationships">\n'
            '  <c:date1904 val="0"/>\n'
            '  <c:roundedCorners val="0"/>\n'
            "  <c:chart>\n"
            '    <c:autoTitleDeleted val="0"/>\n'
            "    <c:plotArea>\n"
            "      <c:layout/>\n"
            "      <c:doughnutChart>\n"
            '        <c:varyColors val="1"/>\n'
            "{ser_xml}"
            "        <c:dLbls>\n"
            '          <c:showLegendKey val="0"/>\n'
            '          <c:showVal val="0"/>\n'
            '          <c:showCatName val="0"/>\n'
            '          <c:showSerName val="0"/>\n'
            '          <c:showPercent val="0"/>\n'
            '          <c:showBubbleSize val="0"/>\n'
            '          <c:showLeaderLines val="1"/>\n'
            "        </c:dLbls>\n"
            '        <c:firstSliceAng val="0"/>\n'
            '        <c:holeSize val="50"/>\n'
            "      </c:doughnutChart>\n"
            "    </c:plotArea>\n"
            "    <c:legend>\n"
            '      <c:legendPos val="r"/>\n'
            "      <c:layout/>\n"
            '      <c:overlay val="0"/>\n'
            "    </c:legend>\n"
            '    <c:plotVisOnly val="1"/>\n'
            '    <c:dispBlanksAs val="gap"/>\n'
            '    <c:showDLblsOverMax val="0"/>\n'
            "  </c:chart>\n"
            "  <c:txPr>\n"
            "    <a:bodyPr/>\n"
            "    <a:lstStyle/>\n"
            "    <a:p>\n"
            "      <a:pPr>\n"
            '        <a:defRPr sz="1800"/>\n'
            "      </a:pPr>\n"
            "      <a:endParaRPr/>\n"
            "    </a:p>\n"
            "  </c:txPr>\n"
            "</c:chartSpace>\n"
        ).format(**{"ser_xml": self._ser_xml})

    @property
    def _explosion_xml(self):
        if self._chart_type == XL_CHART_TYPE.DOUGHNUT_EXPLODED:
            return '          <c:explosion val="25"/>\n'
        return ""

    @property
    def _ser_xml(self):
        xml = ""
        for series in self._chart_data:
            xml_writer = _CategorySeriesXmlWriter(series)
            xml += (
                "        <c:ser>\n"
                '          <c:idx val="{ser_idx}"/>\n'
                '          <c:order val="{ser_order}"/>\n'
                "{tx_xml}"
                "{explosion_xml}"
                "{cat_xml}"
                "{val_xml}"
                "        </c:ser>\n"
            ).format(
                **{
                    "ser_idx": series.index,
                    "ser_order": series.index,
                    "tx_xml": xml_writer.tx_xml,
                    "explosion_xml": self._explosion_xml,
                    "cat_xml": xml_writer.cat_xml,
                    "val_xml": xml_writer.val_xml,
                }
            )
        return xml


class _LineChartXmlWriter(_BaseChartXmlWriter):
    """
    Provides specialized methods particular to the ``<c:lineChart>`` element.
    """

    @property
    def xml(self):
        return (
            "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
            '<c:chartSpace xmlns:c="http://schemas.openxmlformats.org/drawin'
            'gml/2006/chart" xmlns:a="http://schemas.openxmlformats.org/draw'
            'ingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/off'
            'iceDocument/2006/relationships">\n'
            '  <c:date1904 val="0"/>\n'
            "  <c:chart>\n"
            '    <c:autoTitleDeleted val="0"/>\n'
            "    <c:plotArea>\n"
            "      <c:lineChart>\n"
            "{grouping_xml}"
            '        <c:varyColors val="0"/>\n'
            "{ser_xml}"
            '        <c:marker val="1"/>\n'
            '        <c:smooth val="0"/>\n'
            '        <c:axId val="2118791784"/>\n'
            '        <c:axId val="2140495176"/>\n'
            "      </c:lineChart>\n"
            "{cat_ax_xml}"
            "      <c:valAx>\n"
            '        <c:axId val="2140495176"/>\n'
            "        <c:scaling/>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="l"/>\n'
            "        <c:majorGridlines/>\n"
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="2118791784"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            "      </c:valAx>\n"
            "    </c:plotArea>\n"
            "    <c:legend>\n"
            '      <c:legendPos val="r"/>\n'
            "      <c:layout/>\n"
            '      <c:overlay val="0"/>\n'
            "    </c:legend>\n"
            '    <c:plotVisOnly val="1"/>\n'
            '    <c:dispBlanksAs val="gap"/>\n'
            '    <c:showDLblsOverMax val="0"/>\n'
            "  </c:chart>\n"
            "  <c:txPr>\n"
            "    <a:bodyPr/>\n"
            "    <a:lstStyle/>\n"
            "    <a:p>\n"
            "      <a:pPr>\n"
            '        <a:defRPr sz="1800"/>\n'
            "      </a:pPr>\n"
            '      <a:endParaRPr lang="en-US"/>\n'
            "    </a:p>\n"
            "  </c:txPr>\n"
            "</c:chartSpace>\n"
        ).format(
            **{
                "grouping_xml": self._grouping_xml,
                "ser_xml": self._ser_xml,
                "cat_ax_xml": self._cat_ax_xml,
            }
        )

    @property
    def _cat_ax_xml(self):
        categories = self._chart_data.categories

        if categories.are_dates:
            return (
                "      <c:dateAx>\n"
                '        <c:axId val="2118791784"/>\n'
                "        <c:scaling>\n"
                '          <c:orientation val="minMax"/>\n'
                "        </c:scaling>\n"
                '        <c:delete val="0"/>\n'
                '        <c:axPos val="b"/>\n'
                '        <c:numFmt formatCode="{nf}" sourceLinked="1"/>\n'
                '        <c:majorTickMark val="out"/>\n'
                '        <c:minorTickMark val="none"/>\n'
                '        <c:tickLblPos val="nextTo"/>\n'
                '        <c:crossAx val="2140495176"/>\n'
                '        <c:crosses val="autoZero"/>\n'
                '        <c:auto val="1"/>\n'
                '        <c:lblOffset val="100"/>\n'
                '        <c:baseTimeUnit val="days"/>\n'
                "      </c:dateAx>\n"
            ).format(**{"nf": categories.number_format})

        return (
            "      <c:catAx>\n"
            '        <c:axId val="2118791784"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="b"/>\n'
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="2140495176"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:auto val="1"/>\n'
            '        <c:lblAlgn val="ctr"/>\n'
            '        <c:lblOffset val="100"/>\n'
            '        <c:noMultiLvlLbl val="0"/>\n'
            "      </c:catAx>\n"
        )

    @property
    def _grouping_xml(self):
        XL = XL_CHART_TYPE
        standard_types = (XL.LINE, XL.LINE_MARKERS)
        stacked_types = (XL.LINE_STACKED, XL.LINE_MARKERS_STACKED)
        percentStacked_types = (XL.LINE_STACKED_100, XL.LINE_MARKERS_STACKED_100)
        if self._chart_type in standard_types:
            return '        <c:grouping val="standard"/>\n'
        elif self._chart_type in stacked_types:
            return '        <c:grouping val="stacked"/>\n'
        elif self._chart_type in percentStacked_types:
            return '        <c:grouping val="percentStacked"/>\n'
        raise NotImplementedError(
            "no _grouping_xml() for chart type %s" % self._chart_type
        )

    @property
    def _marker_xml(self):
        XL = XL_CHART_TYPE
        no_marker_types = (XL.LINE, XL.LINE_STACKED, XL.LINE_STACKED_100)
        if self._chart_type in no_marker_types:
            return (
                "          <c:marker>\n"
                '            <c:symbol val="none"/>\n'
                "          </c:marker>\n"
            )
        return ""

    @property
    def _ser_xml(self):
        xml = ""
        for series in self._chart_data:
            xml_writer = _CategorySeriesXmlWriter(series)
            xml += (
                "        <c:ser>\n"
                '          <c:idx val="{ser_idx}"/>\n'
                '          <c:order val="{ser_order}"/>\n'
                "{tx_xml}"
                "{marker_xml}"
                "{cat_xml}"
                "{val_xml}"
                '          <c:smooth val="0"/>\n'
                "        </c:ser>\n"
            ).format(
                **{
                    "ser_idx": series.index,
                    "ser_order": series.index,
                    "tx_xml": xml_writer.tx_xml,
                    "marker_xml": self._marker_xml,
                    "cat_xml": xml_writer.cat_xml,
                    "val_xml": xml_writer.val_xml,
                }
            )
        return xml


class _PieChartXmlWriter(_BaseChartXmlWriter):
    """
    Provides specialized methods particular to the ``<c:pieChart>`` element.
    """

    @property
    def xml(self):
        return (
            "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
            '<c:chartSpace xmlns:c="http://schemas.openxmlformats.org/drawin'
            'gml/2006/chart" xmlns:a="http://schemas.openxmlformats.org/draw'
            'ingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/off'
            'iceDocument/2006/relationships">\n'
            "  <c:chart>\n"
            '    <c:autoTitleDeleted val="0"/>\n'
            "    <c:plotArea>\n"
            "      <c:pieChart>\n"
            '        <c:varyColors val="1"/>\n'
            "{ser_xml}"
            "      </c:pieChart>\n"
            "    </c:plotArea>\n"
            '    <c:dispBlanksAs val="gap"/>\n'
            "  </c:chart>\n"
            "  <c:txPr>\n"
            "    <a:bodyPr/>\n"
            "    <a:lstStyle/>\n"
            "    <a:p>\n"
            "      <a:pPr>\n"
            '        <a:defRPr sz="1800"/>\n'
            "      </a:pPr>\n"
            '      <a:endParaRPr lang="en-US"/>\n'
            "    </a:p>\n"
            "  </c:txPr>\n"
            "</c:chartSpace>\n"
        ).format(**{"ser_xml": self._ser_xml})

    @property
    def _explosion_xml(self):
        if self._chart_type == XL_CHART_TYPE.PIE_EXPLODED:
            return '          <c:explosion val="25"/>\n'
        return ""

    @property
    def _ser_xml(self):
        xml_writer = _CategorySeriesXmlWriter(self._chart_data[0])
        xml = (
            "        <c:ser>\n"
            '          <c:idx val="0"/>\n'
            '          <c:order val="0"/>\n'
            "{tx_xml}"
            "{explosion_xml}"
            "{cat_xml}"
            "{val_xml}"
            "        </c:ser>\n"
        ).format(
            **{
                "tx_xml": xml_writer.tx_xml,
                "explosion_xml": self._explosion_xml,
                "cat_xml": xml_writer.cat_xml,
                "val_xml": xml_writer.val_xml,
            }
        )
        return xml


class _RadarChartXmlWriter(_BaseChartXmlWriter):
    """
    Generates XML for the ``<c:radarChart>`` element.
    """

    @property
    def xml(self):
        return (
            "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
            '<c:chartSpace xmlns:c="http://schemas.openxmlformats.org/drawin'
            'gml/2006/chart" xmlns:a="http://schemas.openxmlformats.org/draw'
            'ingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/off'
            'iceDocument/2006/relationships">\n'
            '  <c:date1904 val="0"/>\n'
            '  <c:roundedCorners val="0"/>\n'
            '  <mc:AlternateContent xmlns:mc="http://schemas.openxmlformats.'
            'org/markup-compatibility/2006">\n'
            '    <mc:Choice xmlns:c14="http://schemas.microsoft.com/office/d'
            'rawing/2007/8/2/chart" Requires="c14">\n'
            '      <c14:style val="118"/>\n'
            "    </mc:Choice>\n"
            "    <mc:Fallback>\n"
            '      <c:style val="18"/>\n'
            "    </mc:Fallback>\n"
            "  </mc:AlternateContent>\n"
            "  <c:chart>\n"
            '    <c:autoTitleDeleted val="0"/>\n'
            "    <c:plotArea>\n"
            "      <c:layout/>\n"
            "      <c:radarChart>\n"
            '        <c:radarStyle val="{radar_style}"/>\n'
            '        <c:varyColors val="0"/>\n'
            "{ser_xml}"
            '        <c:axId val="2073612648"/>\n'
            '        <c:axId val="-2112772216"/>\n'
            "      </c:radarChart>\n"
            "      <c:catAx>\n"
            '        <c:axId val="2073612648"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="b"/>\n'
            "        <c:majorGridlines/>\n"
            '        <c:numFmt formatCode="m/d/yy" sourceLinked="1"/>\n'
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="-2112772216"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:auto val="1"/>\n'
            '        <c:lblAlgn val="ctr"/>\n'
            '        <c:lblOffset val="100"/>\n'
            '        <c:noMultiLvlLbl val="0"/>\n'
            "      </c:catAx>\n"
            "      <c:valAx>\n"
            '        <c:axId val="-2112772216"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="l"/>\n'
            "        <c:majorGridlines/>\n"
            '        <c:numFmt formatCode="General" sourceLinked="1"/>\n'
            '        <c:majorTickMark val="cross"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="2073612648"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:crossBetween val="between"/>\n'
            "      </c:valAx>\n"
            "    </c:plotArea>\n"
            '    <c:plotVisOnly val="1"/>\n'
            '    <c:dispBlanksAs val="gap"/>\n'
            '    <c:showDLblsOverMax val="0"/>\n'
            "  </c:chart>\n"
            "  <c:txPr>\n"
            "    <a:bodyPr/>\n"
            "    <a:lstStyle/>\n"
            "    <a:p>\n"
            "      <a:pPr>\n"
            '        <a:defRPr sz="1800"/>\n'
            "      </a:pPr>\n"
            '      <a:endParaRPr lang="en-US"/>\n'
            "    </a:p>\n"
            "  </c:txPr>\n"
            "</c:chartSpace>\n"
        ).format(**{"radar_style": self._radar_style, "ser_xml": self._ser_xml})

    @property
    def _marker_xml(self):
        if self._chart_type == XL_CHART_TYPE.RADAR:
            return (
                "          <c:marker>\n"
                '            <c:symbol val="none"/>\n'
                "          </c:marker>\n"
            )
        return ""

    @property
    def _radar_style(self):
        if self._chart_type == XL_CHART_TYPE.RADAR_FILLED:
            return "filled"
        return "marker"

    @property
    def _ser_xml(self):
        xml = ""
        for series in self._chart_data:
            xml_writer = _CategorySeriesXmlWriter(series)
            xml += (
                "        <c:ser>\n"
                '          <c:idx val="{ser_idx}"/>\n'
                '          <c:order val="{ser_order}"/>\n'
                "{tx_xml}"
                "{marker_xml}"
                "{cat_xml}"
                "{val_xml}"
                '          <c:smooth val="0"/>\n'
                "        </c:ser>\n"
            ).format(
                **{
                    "ser_idx": series.index,
                    "ser_order": series.index,
                    "tx_xml": xml_writer.tx_xml,
                    "marker_xml": self._marker_xml,
                    "cat_xml": xml_writer.cat_xml,
                    "val_xml": xml_writer.val_xml,
                }
            )
        return xml


class _XyChartXmlWriter(_BaseChartXmlWriter):
    """
    Generates XML for the ``<c:scatterChart>`` element.
    """

    @property
    def xml(self):
        xml = (
            "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
            '<c:chartSpace xmlns:c="http://schemas.openxmlformats.org/drawin'
            'gml/2006/chart" xmlns:a="http://schemas.openxmlformats.org/draw'
            'ingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/off'
            'iceDocument/2006/relationships">\n'
            "  <c:chart>\n"
            "    <c:plotArea>\n"
            "      <c:scatterChart>\n"
            '        <c:scatterStyle val="%s"/>\n'
            '        <c:varyColors val="0"/>\n'
            "%s"
            '        <c:axId val="-2128940872"/>\n'
            '        <c:axId val="-2129643912"/>\n'
            "      </c:scatterChart>\n"
            "      <c:valAx>\n"
            '        <c:axId val="-2128940872"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="b"/>\n'
            '        <c:numFmt formatCode="General" sourceLinked="1"/>\n'
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="-2129643912"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:crossBetween val="midCat"/>\n'
            "      </c:valAx>\n"
            "      <c:valAx>\n"
            '        <c:axId val="-2129643912"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="l"/>\n'
            "        <c:majorGridlines/>\n"
            '        <c:numFmt formatCode="General" sourceLinked="1"/>\n'
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="-2128940872"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:crossBetween val="midCat"/>\n'
            "      </c:valAx>\n"
            "    </c:plotArea>\n"
            "    <c:legend>\n"
            '      <c:legendPos val="r"/>\n'
            "      <c:layout/>\n"
            '      <c:overlay val="0"/>\n'
            "    </c:legend>\n"
            '    <c:plotVisOnly val="1"/>\n'
            '    <c:dispBlanksAs val="gap"/>\n'
            '    <c:showDLblsOverMax val="0"/>\n'
            "  </c:chart>\n"
            "  <c:txPr>\n"
            "    <a:bodyPr/>\n"
            "    <a:lstStyle/>\n"
            "    <a:p>\n"
            "      <a:pPr>\n"
            '        <a:defRPr sz="1800"/>\n'
            "      </a:pPr>\n"
            '      <a:endParaRPr lang="en-US"/>\n'
            "    </a:p>\n"
            "  </c:txPr>\n"
            "</c:chartSpace>\n"
        ) % (self._scatterStyle_val, self._ser_xml)
        return xml

    @property
    def _marker_xml(self):
        no_marker_types = (
            XL_CHART_TYPE.XY_SCATTER_LINES_NO_MARKERS,
            XL_CHART_TYPE.XY_SCATTER_SMOOTH_NO_MARKERS,
        )
        if self._chart_type in no_marker_types:
            return (
                "          <c:marker>\n"
                '            <c:symbol val="none"/>\n'
                "          </c:marker>\n"
            )
        return ""

    @property
    def _scatterStyle_val(self):
        smooth_types = (
            XL_CHART_TYPE.XY_SCATTER_SMOOTH,
            XL_CHART_TYPE.XY_SCATTER_SMOOTH_NO_MARKERS,
        )
        if self._chart_type in smooth_types:
            return "smoothMarker"
        return "lineMarker"

    @property
    def _ser_xml(self):
        xml = ""
        for series in self._chart_data:
            xml_writer = _XySeriesXmlWriter(series)
            xml += (
                "        <c:ser>\n"
                '          <c:idx val="{ser_idx}"/>\n'
                '          <c:order val="{ser_order}"/>\n'
                "{tx_xml}"
                "{spPr_xml}"
                "{marker_xml}"
                "{xVal_xml}"
                "{yVal_xml}"
                '          <c:smooth val="0"/>\n'
                "        </c:ser>\n"
            ).format(
                **{
                    "ser_idx": series.index,
                    "ser_order": series.index,
                    "tx_xml": xml_writer.tx_xml,
                    "spPr_xml": self._spPr_xml,
                    "marker_xml": self._marker_xml,
                    "xVal_xml": xml_writer.xVal_xml,
                    "yVal_xml": xml_writer.yVal_xml,
                }
            )
        return xml

    @property
    def _spPr_xml(self):
        if self._chart_type == XL_CHART_TYPE.XY_SCATTER:
            return (
                "          <c:spPr>\n"
                '            <a:ln w="47625">\n'
                "              <a:noFill/>\n"
                "            </a:ln>\n"
                "          </c:spPr>\n"
            )
        return ""


class _BubbleChartXmlWriter(_XyChartXmlWriter):
    """
    Provides specialized methods particular to the ``<c:bubbleChart>``
    element.
    """

    @property
    def xml(self):
        xml = (
            "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
            '<c:chartSpace xmlns:c="http://schemas.openxmlformats.org/drawin'
            'gml/2006/chart" xmlns:a="http://schemas.openxmlformats.org/draw'
            'ingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/off'
            'iceDocument/2006/relationships">\n'
            "  <c:chart>\n"
            '    <c:autoTitleDeleted val="0"/>\n'
            "    <c:plotArea>\n"
            "      <c:layout/>\n"
            "      <c:bubbleChart>\n"
            '        <c:varyColors val="0"/>\n'
            "%s"
            "        <c:dLbls>\n"
            '          <c:showLegendKey val="0"/>\n'
            '          <c:showVal val="0"/>\n'
            '          <c:showCatName val="0"/>\n'
            '          <c:showSerName val="0"/>\n'
            '          <c:showPercent val="0"/>\n'
            '          <c:showBubbleSize val="0"/>\n'
            "        </c:dLbls>\n"
            '        <c:bubbleScale val="100"/>\n'
            '        <c:showNegBubbles val="0"/>\n'
            '        <c:axId val="-2115720072"/>\n'
            '        <c:axId val="-2115723560"/>\n'
            "      </c:bubbleChart>\n"
            "      <c:valAx>\n"
            '        <c:axId val="-2115720072"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="b"/>\n'
            '        <c:numFmt formatCode="General" sourceLinked="1"/>\n'
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="-2115723560"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:crossBetween val="midCat"/>\n'
            "      </c:valAx>\n"
            "      <c:valAx>\n"
            '        <c:axId val="-2115723560"/>\n'
            "        <c:scaling>\n"
            '          <c:orientation val="minMax"/>\n'
            "        </c:scaling>\n"
            '        <c:delete val="0"/>\n'
            '        <c:axPos val="l"/>\n'
            "        <c:majorGridlines/>\n"
            '        <c:numFmt formatCode="General" sourceLinked="1"/>\n'
            '        <c:majorTickMark val="out"/>\n'
            '        <c:minorTickMark val="none"/>\n'
            '        <c:tickLblPos val="nextTo"/>\n'
            '        <c:crossAx val="-2115720072"/>\n'
            '        <c:crosses val="autoZero"/>\n'
            '        <c:crossBetween val="midCat"/>\n'
            "      </c:valAx>\n"
            "    </c:plotArea>\n"
            "    <c:legend>\n"
            '      <c:legendPos val="r"/>\n'
            "      <c:layout/>\n"
            '      <c:overlay val="0"/>\n'
            "    </c:legend>\n"
            '    <c:plotVisOnly val="1"/>\n'
            '    <c:dispBlanksAs val="gap"/>\n'
            '    <c:showDLblsOverMax val="0"/>\n'
            "  </c:chart>\n"
            "  <c:txPr>\n"
            "    <a:bodyPr/>\n"
            "    <a:lstStyle/>\n"
            "    <a:p>\n"
            "      <a:pPr>\n"
            '        <a:defRPr sz="1800"/>\n'
            "      </a:pPr>\n"
            '      <a:endParaRPr lang="en-US"/>\n'
            "    </a:p>\n"
            "  </c:txPr>\n"
            "</c:chartSpace>\n"
        ) % self._ser_xml
        return xml

    @property
    def _bubble3D_val(self):
        if self._chart_type == XL_CHART_TYPE.BUBBLE_THREE_D_EFFECT:
            return "1"
        return "0"

    @property
    def _ser_xml(self):
        xml = ""
        for series in self._chart_data:
            xml_writer = _BubbleSeriesXmlWriter(series)
            xml += (
                "        <c:ser>\n"
                '          <c:idx val="{ser_idx}"/>\n'
                '          <c:order val="{ser_order}"/>\n'
                "{tx_xml}"
                '          <c:invertIfNegative val="0"/>\n'
                "{xVal_xml}"
                "{yVal_xml}"
                "{bubbleSize_xml}"
                '          <c:bubble3D val="{bubble3D_val}"/>\n'
                "        </c:ser>\n"
            ).format(
                **{
                    "ser_idx": series.index,
                    "ser_order": series.index,
                    "tx_xml": xml_writer.tx_xml,
                    "xVal_xml": xml_writer.xVal_xml,
                    "yVal_xml": xml_writer.yVal_xml,
                    "bubbleSize_xml": xml_writer.bubbleSize_xml,
                    "bubble3D_val": self._bubble3D_val,
                }
            )
        return xml


class _CategorySeriesXmlWriter(_BaseSeriesXmlWriter):
    """
    Generates XML snippets particular to a category chart series.
    """

    @property
    def cat(self):
        """
        Return the ``<c:cat>`` element XML for this series, as an oxml
        element.
        """
        categories = self._series.categories

        if categories.are_numeric:
            return parse_xml(
                self._numRef_cat_tmpl.format(
                    **{
                        "wksht_ref": self._series.categories_ref,
                        "number_format": categories.number_format,
                        "cat_count": categories.leaf_count,
                        "cat_pt_xml": self._cat_num_pt_xml,
                        "nsdecls": " %s" % nsdecls("c"),
                    }
                )
            )

        if categories.depth == 1:
            return parse_xml(
                self._cat_tmpl.format(
                    **{
                        "wksht_ref": self._series.categories_ref,
                        "cat_count": categories.leaf_count,
                        "cat_pt_xml": self._cat_pt_xml,
                        "nsdecls": " %s" % nsdecls("c"),
                    }
                )
            )

        return parse_xml(
            self._multiLvl_cat_tmpl.format(
                **{
                    "wksht_ref": self._series.categories_ref,
                    "cat_count": categories.leaf_count,
                    "lvl_xml": self._lvl_xml(categories),
                    "nsdecls": " %s" % nsdecls("c"),
                }
            )
        )

    @property
    def cat_xml(self):
        """
        The unicode XML snippet for the ``<c:cat>`` element for this series,
        containing the category labels and spreadsheet reference.
        """
        categories = self._series.categories

        if categories.are_numeric:
            return self._numRef_cat_tmpl.format(
                **{
                    "wksht_ref": self._series.categories_ref,
                    "number_format": categories.number_format,
                    "cat_count": categories.leaf_count,
                    "cat_pt_xml": self._cat_num_pt_xml,
                    "nsdecls": "",
                }
            )

        if categories.depth == 1:
            return self._cat_tmpl.format(
                **{
                    "wksht_ref": self._series.categories_ref,
                    "cat_count": categories.leaf_count,
                    "cat_pt_xml": self._cat_pt_xml,
                    "nsdecls": "",
                }
            )

        return self._multiLvl_cat_tmpl.format(
            **{
                "wksht_ref": self._series.categories_ref,
                "cat_count": categories.leaf_count,
                "lvl_xml": self._lvl_xml(categories),
                "nsdecls": "",
            }
        )

    @property
    def val(self):
        """
        The ``<c:val>`` XML for this series, as an oxml element.
        """
        xml = self._val_tmpl.format(
            **{
                "nsdecls": " %s" % nsdecls("c"),
                "values_ref": self._series.values_ref,
                "number_format": self._series.number_format,
                "val_count": len(self._series),
                "val_pt_xml": self._val_pt_xml,
            }
        )
        return parse_xml(xml)

    @property
    def val_xml(self):
        """
        Return the unicode XML snippet for the ``<c:val>`` element describing
        this series, containing the series values and their spreadsheet range
        reference.
        """
        return self._val_tmpl.format(
            **{
                "nsdecls": "",
                "values_ref": self._series.values_ref,
                "number_format": self._series.number_format,
                "val_count": len(self._series),
                "val_pt_xml": self._val_pt_xml,
            }
        )

    @property
    def _cat_num_pt_xml(self):
        """
        The unicode XML snippet for the ``<c:pt>`` elements when category
        labels are numeric (including date type).
        """
        xml = ""
        for idx, category in enumerate(self._series.categories):
            xml += (
                '                <c:pt idx="{cat_idx}">\n'
                "                  <c:v>{cat_lbl_str}</c:v>\n"
                "                </c:pt>\n"
            ).format(
                **{
                    "cat_idx": idx,
                    "cat_lbl_str": category.numeric_str_val(self._date_1904),
                }
            )
        return xml

    @property
    def _cat_pt_xml(self):
        """
        The unicode XML snippet for the ``<c:pt>`` elements containing the
        category names for this series.
        """
        xml = ""
        for idx, category in enumerate(self._series.categories):
            xml += (
                '                <c:pt idx="{cat_idx}">\n'
                "                  <c:v>{cat_label}</c:v>\n"
                "                </c:pt>\n"
            ).format(
                **{"cat_idx": idx, "cat_label": escape(to_unicode(category.label))}
            )
        return xml

    @property
    def _cat_tmpl(self):
        """
        The template for the ``<c:cat>`` element for this series, containing
        the category labels and spreadsheet reference.
        """
        return (
            "          <c:cat{nsdecls}>\n"
            "            <c:strRef>\n"
            "              <c:f>{wksht_ref}</c:f>\n"
            "              <c:strCache>\n"
            '                <c:ptCount val="{cat_count}"/>\n'
            "{cat_pt_xml}"
            "              </c:strCache>\n"
            "            </c:strRef>\n"
            "          </c:cat>\n"
        )

    def _lvl_xml(self, categories):
        """
        The unicode XML snippet for the ``<c:lvl>`` elements containing
        multi-level category names.
        """

        def lvl_pt_xml(level):
            xml = ""
            for idx, name in level:
                xml += (
                    '                  <c:pt idx="%d">\n'
                    "                    <c:v>%s</c:v>\n"
                    "                  </c:pt>\n"
                ) % (idx, escape("%s" % name))
            return xml

        xml = ""
        for level in categories.levels:
            xml += (
                "                <c:lvl>\n" "{lvl_pt_xml}" "                </c:lvl>\n"
            ).format(**{"lvl_pt_xml": lvl_pt_xml(level)})
        return xml

    @property
    def _multiLvl_cat_tmpl(self):
        """
        The template for the ``<c:cat>`` element for this series when there
        are multi-level (nested) categories.
        """
        return (
            "          <c:cat{nsdecls}>\n"
            "            <c:multiLvlStrRef>\n"
            "              <c:f>{wksht_ref}</c:f>\n"
            "              <c:multiLvlStrCache>\n"
            '                <c:ptCount val="{cat_count}"/>\n'
            "{lvl_xml}"
            "              </c:multiLvlStrCache>\n"
            "            </c:multiLvlStrRef>\n"
            "          </c:cat>\n"
        )

    @property
    def _numRef_cat_tmpl(self):
        """
        The template for the ``<c:cat>`` element for this series when the
        labels are numeric (or date) values.
        """
        return (
            "          <c:cat{nsdecls}>\n"
            "            <c:numRef>\n"
            "              <c:f>{wksht_ref}</c:f>\n"
            "              <c:numCache>\n"
            "                <c:formatCode>{number_format}</c:formatCode>\n"
            '                <c:ptCount val="{cat_count}"/>\n'
            "{cat_pt_xml}"
            "              </c:numCache>\n"
            "            </c:numRef>\n"
            "          </c:cat>\n"
        )

    @property
    def _val_pt_xml(self):
        """
        The unicode XML snippet containing the ``<c:pt>`` elements containing
        the values for this series.
        """
        xml = ""
        for idx, value in enumerate(self._series.values):
            if value is None:
                continue
            xml += (
                '                <c:pt idx="{val_idx:d}">\n'
                "                  <c:v>{value}</c:v>\n"
                "                </c:pt>\n"
            ).format(**{"val_idx": idx, "value": value})
        return xml

    @property
    def _val_tmpl(self):
        """
        The template for the ``<c:val>`` element for this series, containing
        the series values and their spreadsheet range reference.
        """
        return (
            "          <c:val{nsdecls}>\n"
            "            <c:numRef>\n"
            "              <c:f>{values_ref}</c:f>\n"
            "              <c:numCache>\n"
            "                <c:formatCode>{number_format}</c:formatCode>\n"
            '                <c:ptCount val="{val_count}"/>\n'
            "{val_pt_xml}"
            "              </c:numCache>\n"
            "            </c:numRef>\n"
            "          </c:val>\n"
        )


class _XySeriesXmlWriter(_BaseSeriesXmlWriter):
    """
    Generates XML snippets particular to an XY series.
    """

    @property
    def xVal(self):
        """
        Return the ``<c:xVal>`` element for this series as an oxml element.
        This element contains the X values for this series.
        """
        xml = self._xVal_tmpl.format(
            **{
                "nsdecls": " %s" % nsdecls("c"),
                "numRef_xml": self.numRef_xml(
                    self._series.x_values_ref,
                    self._series.number_format,
                    self._series.x_values,
                ),
            }
        )
        return parse_xml(xml)

    @property
    def xVal_xml(self):
        """
        Return the ``<c:xVal>`` element for this series as unicode text. This
        element contains the X values for this series.
        """
        return self._xVal_tmpl.format(
            **{
                "nsdecls": "",
                "numRef_xml": self.numRef_xml(
                    self._series.x_values_ref,
                    self._series.number_format,
                    self._series.x_values,
                ),
            }
        )

    @property
    def yVal(self):
        """
        Return the ``<c:yVal>`` element for this series as an oxml element.
        This element contains the Y values for this series.
        """
        xml = self._yVal_tmpl.format(
            **{
                "nsdecls": " %s" % nsdecls("c"),
                "numRef_xml": self.numRef_xml(
                    self._series.y_values_ref,
                    self._series.number_format,
                    self._series.y_values,
                ),
            }
        )
        return parse_xml(xml)

    @property
    def yVal_xml(self):
        """
        Return the ``<c:yVal>`` element for this series as unicode text. This
        element contains the Y values for this series.
        """
        return self._yVal_tmpl.format(
            **{
                "nsdecls": "",
                "numRef_xml": self.numRef_xml(
                    self._series.y_values_ref,
                    self._series.number_format,
                    self._series.y_values,
                ),
            }
        )

    @property
    def _xVal_tmpl(self):
        """
        The template for the ``<c:xVal>`` element for this series, containing
        the X values and their spreadsheet range reference.
        """
        return "          <c:xVal{nsdecls}>\n" "{numRef_xml}" "          </c:xVal>\n"

    @property
    def _yVal_tmpl(self):
        """
        The template for the ``<c:yVal>`` element for this series, containing
        the Y values and their spreadsheet range reference.
        """
        return "          <c:yVal{nsdecls}>\n" "{numRef_xml}" "          </c:yVal>\n"


class _BubbleSeriesXmlWriter(_XySeriesXmlWriter):
    """
    Generates XML snippets particular to a bubble chart series.
    """

    @property
    def bubbleSize(self):
        """
        Return the ``<c:bubbleSize>`` element for this series as an oxml
        element. This element contains the bubble size values for this
        series.
        """
        xml = self._bubbleSize_tmpl.format(
            **{
                "nsdecls": " %s" % nsdecls("c"),
                "numRef_xml": self.numRef_xml(
                    self._series.bubble_sizes_ref,
                    self._series.number_format,
                    self._series.bubble_sizes,
                ),
            }
        )
        return parse_xml(xml)

    @property
    def bubbleSize_xml(self):
        """
        Return the ``<c:bubbleSize>`` element for this series as unicode
        text. This element contains the bubble size values for all the
        data points in the chart.
        """
        return self._bubbleSize_tmpl.format(
            **{
                "nsdecls": "",
                "numRef_xml": self.numRef_xml(
                    self._series.bubble_sizes_ref,
                    self._series.number_format,
                    self._series.bubble_sizes,
                ),
            }
        )

    @property
    def _bubbleSize_tmpl(self):
        """
        The template for the ``<c:bubbleSize>`` element for this series,
        containing the bubble size values and their spreadsheet range
        reference.
        """
        return (
            "          <c:bubbleSize{nsdecls}>\n"
            "{numRef_xml}"
            "          </c:bubbleSize>\n"
        )


class _BubbleSeriesXmlRewriter(_BaseSeriesXmlRewriter):
    """
    A series rewriter suitable for bubble charts.
    """

    def _rewrite_ser_data(self, ser, series_data, date_1904):
        """
        Rewrite the ``<c:tx>``, ``<c:cat>`` and ``<c:val>`` child elements
        of *ser* based on the values in *series_data*.
        """
        ser._remove_tx()
        ser._remove_xVal()
        ser._remove_yVal()
        ser._remove_bubbleSize()

        xml_writer = _BubbleSeriesXmlWriter(series_data)

        ser._insert_tx(xml_writer.tx)
        ser._insert_xVal(xml_writer.xVal)
        ser._insert_yVal(xml_writer.yVal)
        ser._insert_bubbleSize(xml_writer.bubbleSize)


class _CategorySeriesXmlRewriter(_BaseSeriesXmlRewriter):
    """
    A series rewriter suitable for category charts.
    """

    def _rewrite_ser_data(self, ser, series_data, date_1904):
        """
        Rewrite the ``<c:tx>``, ``<c:cat>`` and ``<c:val>`` child elements
        of *ser* based on the values in *series_data*.
        """
        ser._remove_tx()
        ser._remove_cat()
        ser._remove_val()

        xml_writer = _CategorySeriesXmlWriter(series_data, date_1904)

        ser._insert_tx(xml_writer.tx)
        ser._insert_cat(xml_writer.cat)
        ser._insert_val(xml_writer.val)


class _XySeriesXmlRewriter(_BaseSeriesXmlRewriter):
    """
    A series rewriter suitable for XY (aka. scatter) charts.
    """

    def _rewrite_ser_data(self, ser, series_data, date_1904):
        """
        Rewrite the ``<c:tx>``, ``<c:xVal>`` and ``<c:yVal>`` child elements
        of *ser* based on the values in *series_data*.
        """
        ser._remove_tx()
        ser._remove_xVal()
        ser._remove_yVal()

        xml_writer = _XySeriesXmlWriter(series_data)

        ser._insert_tx(xml_writer.tx)
        ser._insert_xVal(xml_writer.xVal)
        ser._insert_yVal(xml_writer.yVal)



# encoding: utf-8

"""
Gherkin step implementations for font color features
"""

from __future__ import absolute_import, print_function, unicode_literals

from behave import given, then, when

from pptx import Presentation
from pptx.dml.color import RGBColor
from pptx.enum.dml import MSO_COLOR_TYPE, MSO_THEME_COLOR

from helpers import test_pptx


font_color_pptx_path = test_pptx("font-color")


# given ===================================================


@given("a font with {color_type} color")
def step_given_font_with_color_type(context, color_type):
    context.textbox_idx = {"no": 0, "an RGB": 1, "a theme": 2}[color_type]
    context.prs = Presentation(font_color_pptx_path)
    textbox = context.prs.slides[0].shapes[context.textbox_idx]
    context.font = textbox.text_frame.paragraphs[0].runs[0].font


@given("a font with a color brightness setting of {setting}")
def step_font_with_color_brightness(context, setting):
    textbox_idx = {"no brightness adjustment": 2, "25% darker": 3, "40% lighter": 4}[
        setting
    ]
    context.prs = Presentation(font_color_pptx_path)
    textbox = context.prs.slides[0].shapes[textbox_idx]
    context.font = textbox.text_frame.paragraphs[0].runs[0].font


# when ====================================================


@when("I set the font color brightness to {value}")
def step_set_font_color_brightness(context, value):
    context.font.color.brightness = float(value)


@when("I set the font {color_type} value")
def step_set_font_color_value(context, color_type):
    if color_type == "RGB":
        context.font.color.rgb = RGBColor(0x12, 0x34, 0x56)
    elif color_type == "theme color":
        context.font.color.theme_color = MSO_THEME_COLOR.DARK_1


# then ====================================================


@then("its color value matches its RGB color")
def step_color_value_matches_RGB_color(context):
    assert context.font.color.rgb == RGBColor(255, 102, 0)


@then("its color value matches its theme color")
def step_color_value_matches_theme_color(context):
    assert context.font.color.theme_color == MSO_THEME_COLOR.ACCENT_1


@then("the font's color type is {color_type}")
def step_then_font_color_type_is_value(context, color_type):
    expected_value = {
        "None": None,
        "RGB": MSO_COLOR_TYPE.RGB,
        "theme color": MSO_COLOR_TYPE.SCHEME,
    }[color_type]
    textbox = context.prs.slides[0].shapes[context.textbox_idx]
    font = textbox.text_frame.paragraphs[0].runs[0].font
    assert font.color.type == expected_value


@then("its color brightness value is {value}")
def step_color_brightness_value_matches(context, value):
    assert context.font.color.brightness == float(value)


@then("the font's {color_type} value matches the value I set")
def step_color_type_value_matches(context, color_type):
    textbox = context.prs.slides[0].shapes[context.textbox_idx]
    font = textbox.text_frame.paragraphs[0].runs[0].font
    if color_type == "RGB":
        assert font.color.rgb == RGBColor(0x12, 0x34, 0x56)
    else:
        assert font.color.theme_color == MSO_THEME_COLOR.DARK_1


@then("the font's color brightness is {value}")
def step_color_brightness_matches(context, value):
    textbox = context.prs.slides[0].shapes[context.textbox_idx]
    font = textbox.text_frame.paragraphs[0].runs[0].font
    assert font.color.brightness == float(value)



# encoding: utf-8

"""
Test suite for the docx.shared module
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from pptx.opc.package import XmlPart
from pptx.shared import ElementProxy, ParentedElementProxy

from .unitutil.cxml import element
from .unitutil.mock import instance_mock


class DescribeElementProxy(object):
    def it_raises_on_assign_to_undefined_attr(self):
        element_proxy = ElementProxy(None)
        with pytest.raises(AttributeError):
            element_proxy.foobar = 42

    def it_knows_when_its_equal_to_another_proxy_object(self, eq_fixture):
        proxy, proxy_2, proxy_3, not_a_proxy = eq_fixture

        assert (proxy == proxy_2) is True
        assert (proxy == proxy_3) is False
        assert (proxy == not_a_proxy) is False

        assert (proxy != proxy_2) is False
        assert (proxy != proxy_3) is True
        assert (proxy != not_a_proxy) is True

    def it_knows_its_element(self, element_fixture):
        proxy, element = element_fixture
        assert proxy.element is element

    # fixture --------------------------------------------------------

    @pytest.fixture
    def element_fixture(self):
        p = element("w:p")
        proxy = ElementProxy(p)
        return proxy, p

    @pytest.fixture
    def eq_fixture(self):
        p, q = element("w:p"), element("w:p")
        proxy = ElementProxy(p)
        proxy_2 = ElementProxy(p)
        proxy_3 = ElementProxy(q)
        not_a_proxy = "Foobar"
        return proxy, proxy_2, proxy_3, not_a_proxy


class DescribeParentedElementProxy(object):
    def it_knows_its_parent(self, parent_fixture):
        proxy, parent = parent_fixture
        assert proxy.parent is parent

    def it_knows_its_part(self, part_fixture):
        proxy, part_ = part_fixture
        assert proxy.part is part_

    # fixture --------------------------------------------------------

    @pytest.fixture
    def parent_fixture(self):
        parent = 42
        proxy = ParentedElementProxy(element("w:p"), parent)
        return proxy, parent

    @pytest.fixture
    def part_fixture(self, other_proxy_, part_):
        other_proxy_.part = part_
        proxy = ParentedElementProxy(None, other_proxy_)
        return proxy, part_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def other_proxy_(self, request):
        return instance_mock(request, ParentedElementProxy)

    @pytest.fixture
    def part_(self, request):
        return instance_mock(request, XmlPart)



# encoding: utf-8

"""Slide-related objects, including masters, layouts, and notes."""

from __future__ import absolute_import, division, print_function, unicode_literals

from pptx.dml.fill import FillFormat
from pptx.enum.shapes import PP_PLACEHOLDER
from pptx.shapes.shapetree import (
    LayoutPlaceholders,
    LayoutShapes,
    MasterPlaceholders,
    MasterShapes,
    NotesSlidePlaceholders,
    NotesSlideShapes,
    SlidePlaceholders,
    SlideShapes,
)
from pptx.shared import ElementProxy, ParentedElementProxy, PartElementProxy
from pptx.util import lazyproperty


class _BaseSlide(PartElementProxy):
    """Base class for slide objects, including masters, layouts and notes."""

    __slots__ = ("_background",)

    @lazyproperty
    def background(self):
        """|_Background| object providing slide background properties.

        This property returns a |_Background| object whether or not the
        slide, master, or layout has an explicitly defined background.

        The same |_Background| object is returned on every call for the same
        slide object.
        """
        return _Background(self._element.cSld)

    @property
    def name(self):
        """
        String representing the internal name of this slide. Returns an empty
        string (`''`) if no name is assigned. Assigning an empty string or
        |None| to this property causes any name to be removed.
        """
        return self._element.cSld.name

    @name.setter
    def name(self, value):
        new_value = "" if value is None else value
        self._element.cSld.name = new_value


class _BaseMaster(_BaseSlide):
    """
    Base class for master objects such as |SlideMaster| and |NotesMaster|.
    Provides access to placeholders and regular shapes.
    """

    __slots__ = ("_placeholders", "_shapes")

    @lazyproperty
    def placeholders(self):
        """
        Instance of |MasterPlaceholders| containing sequence of placeholder
        shapes in this master, sorted in *idx* order.
        """
        return MasterPlaceholders(self._element.spTree, self)

    @lazyproperty
    def shapes(self):
        """
        Instance of |MasterShapes| containing sequence of shape objects
        appearing on this slide.
        """
        return MasterShapes(self._element.spTree, self)


class NotesMaster(_BaseMaster):
    """
    Proxy for the notes master XML document. Provides access to shapes, the
    most commonly used of which are placeholders.
    """

    __slots__ = ()


class NotesSlide(_BaseSlide):
    """
    Notes slide object. Provides access to slide notes placeholder and other
    shapes on the notes handout page.
    """

    __slots__ = ("_placeholders", "_shapes")

    def clone_master_placeholders(self, notes_master):
        """
        Selectively add placeholder shape elements from *notes_master* to the
        shapes collection of this notes slide. Z-order of placeholders is
        preserved. Certain placeholders (header, date, footer) are not
        cloned.
        """

        def iter_cloneable_placeholders(notes_master):
            """
            Generate a reference to each placeholder in *notes_master* that
            should be cloned to a notes slide when the a new notes slide is
            created.
            """
            cloneable = (
                PP_PLACEHOLDER.SLIDE_IMAGE,
                PP_PLACEHOLDER.BODY,
                PP_PLACEHOLDER.SLIDE_NUMBER,
            )
            for placeholder in notes_master.placeholders:
                if placeholder.element.ph_type in cloneable:
                    yield placeholder

        shapes = self.shapes
        for placeholder in iter_cloneable_placeholders(notes_master):
            shapes.clone_placeholder(placeholder)

    @property
    def notes_placeholder(self):
        """
        Return the notes placeholder on this notes slide, the shape that
        contains the actual notes text. Return |None| if no notes placeholder
        is present; while this is probably uncommon, it can happen if the
        notes master does not have a body placeholder, or if the notes
        placeholder has been deleted from the notes slide.
        """
        for placeholder in self.placeholders:
            if placeholder.placeholder_format.type == PP_PLACEHOLDER.BODY:
                return placeholder
        return None

    @property
    def notes_text_frame(self):
        """
        Return the text frame of the notes placeholder on this notes slide,
        or |None| if there is no notes placeholder. This is a shortcut to
        accommodate the common case of simply adding "notes" text to the
        notes "page".
        """
        notes_placeholder = self.notes_placeholder
        if notes_placeholder is None:
            return None
        return notes_placeholder.text_frame

    @lazyproperty
    def placeholders(self):
        """
        An instance of |NotesSlidePlaceholders| containing the sequence of
        placeholder shapes in this notes slide.
        """
        return NotesSlidePlaceholders(self.element.spTree, self)

    @lazyproperty
    def shapes(self):
        """
        An instance of |NotesSlideShapes| containing the sequence of shape
        objects appearing on this notes slide.
        """
        return NotesSlideShapes(self._element.spTree, self)


class Slide(_BaseSlide):
    """Slide object. Provides access to shapes and slide-level properties."""

    __slots__ = ("_placeholders", "_shapes")

    @property
    def background(self):
        """|_Background| object providing slide background properties.

        This property returns a |_Background| object whether or not the slide
        overrides the default background or inherits it. Determining which of
        those conditions applies for this slide is accomplished using the
        :attr:`follow_master_background` property.

        The same |_Background| object is returned on every call for the same
        slide object.
        """
        return super(Slide, self).background

    @property
    def follow_master_background(self):
        """|True| if this slide inherits the slide master background.

        Assigning |False| causes background inheritance from the master to be
        interrupted; if there is no custom background for this slide,
        a default background is added. If a custom background already exists
        for this slide, assigning |False| has no effect.

        Assigning |True| causes any custom background for this slide to be
        deleted and inheritance from the master restored.
        """
        return self._element.bg is None

    @property
    def has_notes_slide(self):
        """
        Return True if this slide has a notes slide, False otherwise. A notes
        slide is created by :attr:`.notes_slide` when one doesn't exist; use
        this property to test for a notes slide without the possible side
        effect of creating one.
        """
        return self.part.has_notes_slide

    @property
    def notes_slide(self):
        """
        Return the |NotesSlide| instance for this slide. If the slide does
        not have a notes slide, one is created. The same single instance is
        returned on each call.
        """
        return self.part.notes_slide

    @lazyproperty
    def placeholders(self):
        """
        Instance of |SlidePlaceholders| containing sequence of placeholder
        shapes in this slide.
        """
        return SlidePlaceholders(self._element.spTree, self)

    @lazyproperty
    def shapes(self):
        """
        Instance of |SlideShapes| containing sequence of shape objects
        appearing on this slide.
        """
        return SlideShapes(self._element.spTree, self)

    @property
    def slide_id(self):
        """
        The integer value that uniquely identifies this slide within this
        presentation. The slide id does not change if the position of this
        slide in the slide sequence is changed by adding, rearranging, or
        deleting slides.
        """
        return self.part.slide_id

    @property
    def slide_layout(self):
        """
        |SlideLayout| object this slide inherits appearance from.
        """
        return self.part.slide_layout


class Slides(ParentedElementProxy):
    """
    Sequence of slides belonging to an instance of |Presentation|, having
    list semantics for access to individual slides. Supports indexed access,
    len(), and iteration.
    """

    def __init__(self, sldIdLst, prs):
        super(Slides, self).__init__(sldIdLst, prs)
        self._sldIdLst = sldIdLst

    def __getitem__(self, idx):
        """
        Provide indexed access, (e.g. 'slides[0]').
        """
        try:
            sldId = self._sldIdLst[idx]
        except IndexError:
            raise IndexError("slide index out of range")
        return self.part.related_slide(sldId.rId)

    def __iter__(self):
        """
        Support iteration (e.g. 'for slide in slides:').
        """
        for sldId in self._sldIdLst:
            yield self.part.related_slide(sldId.rId)

    def __len__(self):
        """
        Support len() built-in function (e.g. 'len(slides) == 4').
        """
        return len(self._sldIdLst)

    def add_slide(self, slide_layout):
        """
        Return a newly added slide that inherits layout from *slide_layout*.
        """
        rId, slide = self.part.add_slide(slide_layout)
        slide.shapes.clone_layout_placeholders(slide_layout)
        self._sldIdLst.add_sldId(rId)
        return slide

    def get(self, slide_id, default=None):
        """
        Return the slide identified by integer *slide_id* in this
        presentation, or *default* if not found.
        """
        slide = self.part.get_slide(slide_id)
        if slide is None:
            return default
        return slide

    def index(self, slide):
        """
        Map *slide* to an integer representing its zero-based position in
        this slide collection. Raises |ValueError| on *slide* not present.
        """
        for idx, this_slide in enumerate(self):
            if this_slide == slide:
                return idx
        raise ValueError("%s is not in slide collection" % slide)


class SlideLayout(_BaseSlide):
    """
    Slide layout object. Provides access to placeholders, regular shapes, and
    slide layout-level properties.
    """

    __slots__ = ("_placeholders", "_shapes")

    def iter_cloneable_placeholders(self):
        """
        Generate a reference to each layout placeholder on this slide layout
        that should be cloned to a slide when the layout is applied to that
        slide.
        """
        latent_ph_types = (
            PP_PLACEHOLDER.DATE,
            PP_PLACEHOLDER.FOOTER,
            PP_PLACEHOLDER.SLIDE_NUMBER,
        )
        for ph in self.placeholders:
            if ph.element.ph_type not in latent_ph_types:
                yield ph

    @lazyproperty
    def placeholders(self):
        """
        Instance of |LayoutPlaceholders| containing sequence of placeholder
        shapes in this slide layout, sorted in *idx* order.
        """
        return LayoutPlaceholders(self._element.spTree, self)

    @lazyproperty
    def shapes(self):
        """
        Instance of |LayoutShapes| containing the sequence of shapes
        appearing on this slide layout.
        """
        return LayoutShapes(self._element.spTree, self)

    @property
    def slide_master(self):
        """
        Slide master from which this slide layout inherits properties.
        """
        return self.part.slide_master

    @property
    def used_by_slides(self):
        """Tuple of slide objects based on this slide layout."""
        # ---getting Slides collection requires going around the horn a bit---
        slides = self.part.package.presentation_part.presentation.slides
        return tuple(s for s in slides if s.slide_layout == self)


class SlideLayouts(ParentedElementProxy):
    """Sequence of slide layouts belonging to a slide-master.

    Supports indexed access, len(), iteration, index() and remove().
    """

    __slots__ = ("_sldLayoutIdLst",)

    def __init__(self, sldLayoutIdLst, parent):
        super(SlideLayouts, self).__init__(sldLayoutIdLst, parent)
        self._sldLayoutIdLst = sldLayoutIdLst

    def __getitem__(self, idx):
        """
        Provide indexed access, (e.g. ``slide_layouts[2]``).
        """
        try:
            sldLayoutId = self._sldLayoutIdLst[idx]
        except IndexError:
            raise IndexError("slide layout index out of range")
        return self.part.related_slide_layout(sldLayoutId.rId)

    def __iter__(self):
        """
        Generate a reference to each of the |SlideLayout| instances in the
        collection, in sequence.
        """
        for sldLayoutId in self._sldLayoutIdLst:
            yield self.part.related_slide_layout(sldLayoutId.rId)

    def __len__(self):
        """
        Support len() built-in function (e.g. 'len(slides) == 4').
        """
        return len(self._sldLayoutIdLst)

    def get_by_name(self, name, default=None):
        """Return SlideLayout object having *name* or *default* if not found."""
        for slide_layout in self:
            if slide_layout.name == name:
                return slide_layout
        return default

    def index(self, slide_layout):
        """Return zero-based index of *slide_layout* in this collection.

        Raises ValueError if *slide_layout* is not present in this collection.
        """
        for idx, this_layout in enumerate(self):
            if slide_layout == this_layout:
                return idx
        raise ValueError("layout not in this SlideLayouts collection")

    def remove(self, slide_layout):
        """Remove *slide_layout* from the collection.

        Raises ValueError when *slide_layout* is in use; a slide layout which is the
        basis for one or more slides cannot be removed.
        """
        # ---raise if layout is in use---
        if slide_layout.used_by_slides:
            raise ValueError("cannot remove slide-layout in use by one or more slides")

        # ---target layout is identified by its index in this collection---
        target_idx = self.index(slide_layout)

        # --remove layout from p:sldLayoutIds of its master
        # --this stops layout from showing up, but doesn't remove it from package
        target_sldLayoutId = self._sldLayoutIdLst.sldLayoutId_lst[target_idx]
        self._sldLayoutIdLst.remove(target_sldLayoutId)

        # --drop relationship from master to layout
        # --this removes layout from package, along with everything (only) it refers to,
        # --including images (not used elsewhere) and hyperlinks
        slide_layout.slide_master.part.drop_rel(target_sldLayoutId.rId)


class SlideMaster(_BaseMaster):
    """
    Slide master object. Provides access to slide layouts. Access to
    placeholders, regular shapes, and slide master-level properties is
    inherited from |_BaseMaster|.
    """

    __slots__ = ("_slide_layouts",)

    @lazyproperty
    def slide_layouts(self):
        """|SlideLayouts| object providing access to this slide-master's layouts."""
        return SlideLayouts(self._element.get_or_add_sldLayoutIdLst(), self)


class SlideMasters(ParentedElementProxy):
    """Sequence of |SlideMaster| objects belonging to a presentation.

    Has list access semantics, supporting indexed access, len(), and iteration.
    """

    __slots__ = ("_sldMasterIdLst",)

    def __init__(self, sldMasterIdLst, parent):
        super(SlideMasters, self).__init__(sldMasterIdLst, parent)
        self._sldMasterIdLst = sldMasterIdLst

    def __getitem__(self, idx):
        """
        Provide indexed access, (e.g. ``slide_masters[2]``).
        """
        try:
            sldMasterId = self._sldMasterIdLst[idx]
        except IndexError:
            raise IndexError("slide master index out of range")
        return self.part.related_slide_master(sldMasterId.rId)

    def __iter__(self):
        """
        Generate a reference to each of the |SlideMaster| instances in the
        collection, in sequence.
        """
        for smi in self._sldMasterIdLst:
            yield self.part.related_slide_master(smi.rId)

    def __len__(self):
        """
        Support len() built-in function (e.g. 'len(slide_masters) == 4').
        """
        return len(self._sldMasterIdLst)


class _Background(ElementProxy):
    """Provides access to slide background properties.

    Note that the presence of this object does not by itself imply an
    explicitly-defined background; a slide with an inherited background still
    has a |_Background| object.
    """

    __slots__ = ("_cSld", "_fill")

    def __init__(self, cSld):
        super(_Background, self).__init__(cSld)
        self._cSld = cSld

    @lazyproperty
    def fill(self):
        """|FillFormat| instance for this background.

        This |FillFormat| object is used to interrogate or specify the fill
        of the slide background.

        Note that accessing this property is potentially destructive. A slide
        background can also be specified by a background style reference and
        accessing this property will remove that reference, if present, and
        replace it with NoFill. This is frequently the case for a slide
        master background.

        This is also the case when there is no explicitly defined background
        (background is inherited); merely accessing this property will cause
        the background to be set to NoFill and the inheritance link will be
        interrupted. This is frequently the case for a slide background.

        Of course, if you are accessing this property in order to set the
        fill, then these changes are of no consequence, but the existing
        background cannot be reliably interrogated using this property unless
        you have already established it is an explicit fill.

        If the background is already a fill, then accessing this property
        makes no changes to the current background.
        """
        bgPr = self._cSld.get_or_add_bgPr()
        return FillFormat.from_fill_parent(bgPr)



# encoding: utf-8

"""Autoshape-related objects such as Shape and Adjustment."""

from __future__ import absolute_import, division, print_function, unicode_literals

from numbers import Number

from pptx.dml.fill import FillFormat
from pptx.dml.line import LineFormat
from pptx.enum.shapes import MSO_AUTO_SHAPE_TYPE, MSO_SHAPE_TYPE
from pptx.shapes.base import BaseShape
from pptx.spec import autoshape_types
from pptx.text.text import TextFrame
from pptx.util import lazyproperty


class Adjustment(object):
    """
    An adjustment value for an autoshape.

    An adjustment value corresponds to the position of an adjustment handle on
    an auto shape. Adjustment handles are the small yellow diamond-shaped
    handles that appear on certain auto shapes and allow the outline of the
    shape to be adjusted. For example, a rounded rectangle has an adjustment
    handle that allows the radius of its corner rounding to be adjusted.

    Values are |float| and generally range from 0.0 to 1.0, although the value
    can be negative or greater than 1.0 in certain circumstances.
    """

    def __init__(self, name, def_val, actual=None):
        super(Adjustment, self).__init__()
        self.name = name
        self.def_val = def_val
        self.actual = actual

    @property
    def effective_value(self):
        """
        Read/write |float| representing normalized adjustment value for this
        adjustment. Actual values are a large-ish integer expressed in shape
        coordinates, nominally between 0 and 100,000. The effective value is
        normalized to a corresponding value nominally between 0.0 and 1.0.
        Intuitively this represents the proportion of the width or height of
        the shape at which the adjustment value is located from its starting
        point. For simple shapes such as a rounded rectangle, this intuitive
        correspondence holds. For more complicated shapes and at more extreme
        shape proportions (e.g. width is much greater than height), the value
        can become negative or greater than 1.0.
        """
        raw_value = self.actual
        if raw_value is None:
            raw_value = self.def_val
        return self._normalize(raw_value)

    @effective_value.setter
    def effective_value(self, value):
        if not isinstance(value, Number):
            tmpl = "adjustment value must be numeric, got '%s'"
            raise ValueError(tmpl % value)
        self.actual = self._denormalize(value)

    @staticmethod
    def _denormalize(value):
        """
        Return integer corresponding to normalized *raw_value* on unit basis
        of 100,000. See Adjustment.normalize for additional details.
        """
        return int(value * 100000.0)

    @staticmethod
    def _normalize(raw_value):
        """
        Return normalized value for *raw_value*. A normalized value is a
        |float| between 0.0 and 1.0 for nominal raw values between 0 and
        100,000. Raw values less than 0 and greater than 100,000 are valid
        and return values calculated on the same unit basis of 100,000.
        """
        return raw_value / 100000.0

    @property
    def val(self):
        """
        Denormalized effective value (expressed in shape coordinates),
        suitable for using in the XML.
        """
        return self.actual if self.actual is not None else self.def_val


class AdjustmentCollection(object):
    """
    Sequence of |Adjustment| instances for an auto shape, each representing
    an available adjustment for a shape of its type. Supports ``len()`` and
    indexed access, e.g. ``shape.adjustments[1] = 0.15``.
    """

    def __init__(self, prstGeom):
        super(AdjustmentCollection, self).__init__()
        self._adjustments_ = self._initialized_adjustments(prstGeom)
        self._prstGeom = prstGeom

    def __getitem__(self, key):
        """Provides indexed access, (e.g. 'adjustments[9]')."""
        return self._adjustments_[key].effective_value

    def __setitem__(self, key, value):
        """
        Provides item assignment via an indexed expression, e.g.
        ``adjustments[9] = 999.9``. Causes all adjustment values in
        collection to be written to the XML.
        """
        self._adjustments_[key].effective_value = value
        self._rewrite_guides()

    def _initialized_adjustments(self, prstGeom):
        """
        Return an initialized list of adjustment values based on the contents
        of *prstGeom*
        """
        if prstGeom is None:
            return []
        davs = AutoShapeType.default_adjustment_values(prstGeom.prst)
        adjustments = [Adjustment(name, def_val) for name, def_val in davs]
        self._update_adjustments_with_actuals(adjustments, prstGeom.gd_lst)
        return adjustments

    def _rewrite_guides(self):
        """
        Write ``<a:gd>`` elements to the XML, one for each adjustment value.
        Any existing guide elements are overwritten.
        """
        guides = [(adj.name, adj.val) for adj in self._adjustments_]
        self._prstGeom.rewrite_guides(guides)

    @staticmethod
    def _update_adjustments_with_actuals(adjustments, guides):
        """
        Update |Adjustment| instances in *adjustments* with actual values
        held in *guides*, a list of ``<a:gd>`` elements. Guides with a name
        that does not match an adjustment object are skipped.
        """
        adjustments_by_name = dict((adj.name, adj) for adj in adjustments)
        for gd in guides:
            name = gd.name
            actual = int(gd.fmla[4:])
            try:
                adjustment = adjustments_by_name[name]
            except KeyError:
                continue
            adjustment.actual = actual
        return

    @property
    def _adjustments(self):
        """
        Sequence containing direct references to the |Adjustment| objects
        contained in collection.
        """
        return tuple(self._adjustments_)

    def __len__(self):
        """Implement built-in function len()"""
        return len(self._adjustments_)


class AutoShapeType(object):
    """
    Return an instance of |AutoShapeType| containing metadata for an auto
    shape of type identified by *autoshape_type_id*. Instances are cached, so
    no more than one instance for a particular auto shape type is in memory.

    Instances provide the following attributes:

    .. attribute:: autoshape_type_id

       Integer uniquely identifying this auto shape type. Corresponds to a
       value in ``pptx.constants.MSO`` like ``MSO_SHAPE.ROUNDED_RECTANGLE``.

    .. attribute:: basename

       Base part of shape name for auto shapes of this type, e.g. ``Rounded
       Rectangle`` becomes ``Rounded Rectangle 99`` when the distinguishing
       integer is added to the shape name.

    .. attribute:: prst

       String identifier for this auto shape type used in the ``<a:prstGeom>``
       element.

    .. attribute:: desc

       Informal string description of auto shape.

    """

    _instances = {}

    def __new__(cls, autoshape_type_id):
        """
        Only create new instance on first call for content_type. After that,
        use cached instance.
        """
        # if there's not a matching instance in the cache, create one
        if autoshape_type_id not in cls._instances:
            inst = super(AutoShapeType, cls).__new__(cls)
            cls._instances[autoshape_type_id] = inst
        # return the instance; note that __init__() gets called either way
        return cls._instances[autoshape_type_id]

    def __init__(self, autoshape_type_id):
        """Initialize attributes from constant values in pptx.spec"""
        # skip loading if this instance is from the cache
        if hasattr(self, "_loaded"):
            return
        # raise on bad autoshape_type_id
        if autoshape_type_id not in autoshape_types:
            raise KeyError(
                "no autoshape type with id '%s' in pptx.spec.autoshape_types"
                % autoshape_type_id
            )
        # otherwise initialize new instance
        autoshape_type = autoshape_types[autoshape_type_id]
        self._autoshape_type_id = autoshape_type_id
        self._basename = autoshape_type["basename"]
        self._loaded = True

    @property
    def autoshape_type_id(self):
        """
        MSO_AUTO_SHAPE_TYPE enumeration value for this auto shape type
        """
        return self._autoshape_type_id

    @property
    def basename(self):
        """
        Base of shape name (less the distinguishing integer) for this auto
        shape type
        """
        return self._basename

    @classmethod
    def default_adjustment_values(cls, prst):
        """
        Return sequence of name, value tuples representing the adjustment
        value defaults for the auto shape type identified by *prst*.
        """
        return autoshape_types[prst]["avLst"]

    @property
    def desc(self):
        """Informal description of this auto shape type"""
        return self._desc

    @classmethod
    def id_from_prst(cls, prst):
        """
        Return auto shape id (e.g. ``MSO_SHAPE.RECTANGLE``) corresponding to
        preset geometry keyword *prst*.
        """
        return MSO_AUTO_SHAPE_TYPE.from_xml(prst)

    @property
    def prst(self):
        """
        Preset geometry identifier string for this auto shape. Used in the
        ``prst`` attribute of ``<a:prstGeom>`` element to specify the geometry
        to be used in rendering the shape, for example ``'roundRect'``.
        """
        return MSO_AUTO_SHAPE_TYPE.to_xml(self._autoshape_type_id)


class Shape(BaseShape):
    """A shape that can appear on a slide.

    Corresponds to the ``<p:sp>`` element that can appear in any of the slide-type parts
    (slide, slideLayout, slideMaster, notesPage, notesMaster, handoutMaster).
    """

    def __init__(self, sp, parent):
        super(Shape, self).__init__(sp, parent)
        self._sp = sp

    @lazyproperty
    def adjustments(self):
        """
        Read-only reference to |AdjustmentCollection| instance for this
        shape
        """
        return AdjustmentCollection(self._sp.prstGeom)

    @property
    def auto_shape_type(self):
        """
        Enumeration value identifying the type of this auto shape, like
        ``MSO_SHAPE.ROUNDED_RECTANGLE``. Raises |ValueError| if this shape is
        not an auto shape.
        """
        if not self._sp.is_autoshape:
            raise ValueError("shape is not an auto shape")
        return self._sp.prst

    @lazyproperty
    def fill(self):
        """
        |FillFormat| instance for this shape, providing access to fill
        properties such as fill color.
        """
        return FillFormat.from_fill_parent(self._sp.spPr)

    def get_or_add_ln(self):
        """
        Return the ``<a:ln>`` element containing the line format properties
        XML for this shape.
        """
        return self._sp.get_or_add_ln()

    @property
    def has_text_frame(self):
        """
        |True| if this shape can contain text. Always |True| for an
        AutoShape.
        """
        return True

    @lazyproperty
    def line(self):
        """
        |LineFormat| instance for this shape, providing access to line
        properties such as line color.
        """
        return LineFormat(self)

    @property
    def ln(self):
        """
        The ``<a:ln>`` element containing the line format properties such as
        line color and width. |None| if no ``<a:ln>`` element is present.
        """
        return self._sp.ln

    @property
    def shape_type(self):
        """
        Unique integer identifying the type of this shape, like
        ``MSO_SHAPE_TYPE.TEXT_BOX``.
        """
        if self.is_placeholder:
            return MSO_SHAPE_TYPE.PLACEHOLDER
        if self._sp.has_custom_geometry:
            return MSO_SHAPE_TYPE.FREEFORM
        if self._sp.is_autoshape:
            return MSO_SHAPE_TYPE.AUTO_SHAPE
        if self._sp.is_textbox:
            return MSO_SHAPE_TYPE.TEXT_BOX
        msg = "Shape instance of unrecognized shape type"
        raise NotImplementedError(msg)

    @property
    def text(self):
        """Read/write. Unicode (str in Python 3) representation of shape text.

        The returned string will contain a newline character (``"\\n"``) separating each
        paragraph and a vertical-tab (``"\\v"``) character for each line break (soft
        carriage return) in the shape's text.

        Assignment to *text* replaces all text previously contained in the shape, along
        with any paragraph or font formatting applied to it. A newline character
        (``"\\n"``) in the assigned text causes a new paragraph to be started.
        A vertical-tab (``"\\v"``) character in the assigned text causes a line-break
        (soft carriage-return) to be inserted. (The vertical-tab character appears in
        clipboard text copied from PowerPoint as its encoding of line-breaks.)

        Either bytes (Python 2 str) or unicode (Python 3 str) can be assigned. Bytes can
        be 7-bit ASCII or UTF-8 encoded 8-bit bytes. Bytes values are converted to
        unicode assuming UTF-8 encoding (which also works for ASCII).
        """
        return self.text_frame.text

    @text.setter
    def text(self, text):
        self.text_frame.text = text

    @property
    def text_frame(self):
        """|TextFrame| instance for this shape.

        Contains the text of the shape and provides access to text formatting
        properties.
        """
        txBody = self._element.get_or_add_txBody()
        return TextFrame(txBody, self)



# encoding: utf-8

"""Custom element classes for table-related XML elements"""

from __future__ import absolute_import, division, print_function, unicode_literals

from pptx.enum.text import MSO_VERTICAL_ANCHOR
from pptx.oxml import parse_xml
from pptx.oxml.dml.fill import CT_GradientFillProperties
from pptx.oxml.ns import nsdecls
from pptx.oxml.simpletypes import ST_Coordinate, ST_Coordinate32, XsdBoolean, XsdInt
from pptx.oxml.text import CT_TextBody
from pptx.oxml.xmlchemy import (
    BaseOxmlElement,
    Choice,
    OneAndOnlyOne,
    OptionalAttribute,
    RequiredAttribute,
    ZeroOrMore,
    ZeroOrOne,
    ZeroOrOneChoice,
)
from pptx.util import Emu, lazyproperty


class CT_Table(BaseOxmlElement):
    """`a:tbl` custom element class"""

    _tag_seq = ("a:tblPr", "a:tblGrid", "a:tr")
    tblPr = ZeroOrOne("a:tblPr", successors=_tag_seq[1:])
    tblGrid = OneAndOnlyOne("a:tblGrid")
    tr = ZeroOrMore("a:tr", successors=_tag_seq[3:])
    del _tag_seq

    def add_tr(self, height):
        """
        Return a reference to a newly created <a:tr> child element having its
        ``h`` attribute set to *height*.
        """
        return self._add_tr(h=height)

    @property
    def bandCol(self):
        return self._get_boolean_property("bandCol")

    @bandCol.setter
    def bandCol(self, value):
        self._set_boolean_property("bandCol", value)

    @property
    def bandRow(self):
        return self._get_boolean_property("bandRow")

    @bandRow.setter
    def bandRow(self, value):
        self._set_boolean_property("bandRow", value)

    @property
    def firstCol(self):
        return self._get_boolean_property("firstCol")

    @firstCol.setter
    def firstCol(self, value):
        self._set_boolean_property("firstCol", value)

    @property
    def firstRow(self):
        return self._get_boolean_property("firstRow")

    @firstRow.setter
    def firstRow(self, value):
        self._set_boolean_property("firstRow", value)

    def iter_tcs(self):
        """Generate each `a:tc` element in this tbl.

        tc elements are generated left-to-right, top-to-bottom.
        """
        return (tc for tr in self.tr_lst for tc in tr.tc_lst)

    @property
    def lastCol(self):
        return self._get_boolean_property("lastCol")

    @lastCol.setter
    def lastCol(self, value):
        self._set_boolean_property("lastCol", value)

    @property
    def lastRow(self):
        return self._get_boolean_property("lastRow")

    @lastRow.setter
    def lastRow(self, value):
        self._set_boolean_property("lastRow", value)

    @classmethod
    def new_tbl(cls, rows, cols, width, height, tableStyleId=None):
        """Return a new ``<p:tbl>`` element tree."""
        # working hypothesis is this is the default table style GUID
        if tableStyleId is None:
            tableStyleId = "{5C22544A-7EE6-4342-B048-85BDC9FD1C3A}"

        xml = cls._tbl_tmpl() % (tableStyleId)
        tbl = parse_xml(xml)

        # add specified number of rows and columns
        rowheight = height // rows
        colwidth = width // cols

        for col in range(cols):
            # adjust width of last col to absorb any div error
            if col == cols - 1:
                colwidth = width - ((cols - 1) * colwidth)
            tbl.tblGrid.add_gridCol(width=colwidth)

        for row in range(rows):
            # adjust height of last row to absorb any div error
            if row == rows - 1:
                rowheight = height - ((rows - 1) * rowheight)
            tr = tbl.add_tr(height=rowheight)
            for col in range(cols):
                tr.add_tc()

        return tbl

    def tc(self, row_idx, col_idx):
        """Return `a:tc` element at *row_idx*, *col_idx*."""
        return self.tr_lst[row_idx].tc_lst[col_idx]

    def _get_boolean_property(self, propname):
        """
        Generalized getter for the boolean properties on the ``<a:tblPr>``
        child element. Defaults to False if *propname* attribute is missing
        or ``<a:tblPr>`` element itself is not present.
        """
        tblPr = self.tblPr
        if tblPr is None:
            return False
        propval = getattr(tblPr, propname)
        return {True: True, False: False, None: False}[propval]

    def _set_boolean_property(self, propname, value):
        """
        Generalized setter for boolean properties on the ``<a:tblPr>`` child
        element, setting *propname* attribute appropriately based on *value*.
        If *value* is True, the attribute is set to "1"; a tblPr child
        element is added if necessary. If *value* is False, the *propname*
        attribute is removed if present, allowing its default value of False
        to be its effective value.
        """
        if value not in (True, False):
            raise ValueError(
                "assigned value must be either True or False, got %s" % value
            )
        tblPr = self.get_or_add_tblPr()
        setattr(tblPr, propname, value)

    @classmethod
    def _tbl_tmpl(cls):
        return (
            "<a:tbl %s>\n"
            '  <a:tblPr firstRow="1" bandRow="1">\n'
            "    <a:tableStyleId>%s</a:tableStyleId>\n"
            "  </a:tblPr>\n"
            "  <a:tblGrid/>\n"
            "</a:tbl>" % (nsdecls("a"), "%s")
        )


class CT_TableCell(BaseOxmlElement):
    """`a:tc` custom element class"""

    _tag_seq = ("a:txBody", "a:tcPr", "a:extLst")
    txBody = ZeroOrOne("a:txBody", successors=_tag_seq[1:])
    tcPr = ZeroOrOne("a:tcPr", successors=_tag_seq[2:])
    del _tag_seq

    gridSpan = OptionalAttribute("gridSpan", XsdInt, default=1)
    rowSpan = OptionalAttribute("rowSpan", XsdInt, default=1)
    hMerge = OptionalAttribute("hMerge", XsdBoolean, default=False)
    vMerge = OptionalAttribute("vMerge", XsdBoolean, default=False)

    @property
    def anchor(self):
        """
        String held in ``anchor`` attribute of ``<a:tcPr>`` child element of
        this ``<a:tc>`` element.
        """
        if self.tcPr is None:
            return None
        return self.tcPr.anchor

    @anchor.setter
    def anchor(self, anchor_enum_idx):
        """
        Set value of anchor attribute on ``<a:tcPr>`` child element
        """
        if anchor_enum_idx is None and self.tcPr is None:
            return
        tcPr = self.get_or_add_tcPr()
        tcPr.anchor = anchor_enum_idx

    def append_ps_from(self, spanned_tc):
        """Append `a:p` elements taken from *spanned_tc*.

        Any non-empty paragraph elements in *spanned_tc* are removed and
        appended to the text-frame of this cell. If *spanned_tc* is left with
        no content after this process, a single empty `a:p` element is added
        to ensure the cell is compliant with the spec.
        """
        source_txBody = spanned_tc.get_or_add_txBody()
        target_txBody = self.get_or_add_txBody()

        # ---if source is empty, there's nothing to do---
        if source_txBody.is_empty:
            return

        # ---a single empty paragraph in target is overwritten---
        if target_txBody.is_empty:
            target_txBody.clear_content()

        for p in source_txBody.p_lst:
            target_txBody.append(p)

        # ---neither source nor target can be left without ps---
        source_txBody.unclear_content()
        target_txBody.unclear_content()

    @property
    def col_idx(self):
        """Offset of this cell's column in its table."""
        # ---tc elements come before any others in `a:tr` element---
        return self.getparent().index(self)

    @property
    def is_merge_origin(self):
        """True if cell is top-left in merged cell range."""
        if self.gridSpan > 1 and not self.vMerge:
            return True
        if self.rowSpan > 1 and not self.hMerge:
            return True
        return False

    @property
    def is_spanned(self):
        """True if cell is in merged cell range but not merge origin cell."""
        return self.hMerge or self.vMerge

    @property
    def marT(self):
        """
        Read/write integer top margin value represented in ``marT`` attribute
        of the ``<a:tcPr>`` child element of this ``<a:tc>`` element. If the
        attribute is not present, the default value ``45720`` (0.05 inches)
        is returned for top and bottom; ``91440`` (0.10 inches) is the
        default for left and right. Assigning |None| to any ``marX``
        property clears that attribute from the element, effectively setting
        it to the default value.
        """
        return self._get_marX("marT", 45720)

    @marT.setter
    def marT(self, value):
        self._set_marX("marT", value)

    @property
    def marR(self):
        """
        Right margin value represented in ``marR`` attribute.
        """
        return self._get_marX("marR", 91440)

    @marR.setter
    def marR(self, value):
        self._set_marX("marR", value)

    @property
    def marB(self):
        """
        Bottom margin value represented in ``marB`` attribute.
        """
        return self._get_marX("marB", 45720)

    @marB.setter
    def marB(self, value):
        self._set_marX("marB", value)

    @property
    def marL(self):
        """
        Left margin value represented in ``marL`` attribute.
        """
        return self._get_marX("marL", 91440)

    @marL.setter
    def marL(self, value):
        self._set_marX("marL", value)

    @classmethod
    def new(cls):
        """Return a new `a:tc` element subtree."""
        xml = cls._tc_tmpl()
        tc = parse_xml(xml)
        return tc

    @property
    def row_idx(self):
        """Offset of this cell's row in its table."""
        return self.getparent().row_idx

    @property
    def tbl(self):
        """Table element this cell belongs to."""
        return self.xpath("ancestor::a:tbl")[0]

    @property
    def text(self):
        """str text contained in cell"""
        # ---note this shadows lxml _Element.text---
        txBody = self.txBody
        if txBody is None:
            return ""
        return "\n".join([p.text for p in txBody.p_lst])

    def _get_marX(self, attr_name, default):
        """
        Generalized method to get margin values.
        """
        if self.tcPr is None:
            return Emu(default)
        return Emu(int(self.tcPr.get(attr_name, default)))

    def _new_txBody(self):
        return CT_TextBody.new_a_txBody()

    def _set_marX(self, marX, value):
        """
        Set value of marX attribute on ``<a:tcPr>`` child element. If *marX*
        is |None|, the marX attribute is removed. *marX* is a string, one of
        ``('marL', 'marR', 'marT', 'marB')``.
        """
        if value is None and self.tcPr is None:
            return
        tcPr = self.get_or_add_tcPr()
        setattr(tcPr, marX, value)

    @classmethod
    def _tc_tmpl(cls):
        return (
            "<a:tc %s>\n"
            "  <a:txBody>\n"
            "    <a:bodyPr/>\n"
            "    <a:lstStyle/>\n"
            "    <a:p/>\n"
            "  </a:txBody>\n"
            "  <a:tcPr/>\n"
            "</a:tc>" % nsdecls("a")
        )


class CT_TableCellProperties(BaseOxmlElement):
    """`a:tcPr` custom element class"""

    eg_fillProperties = ZeroOrOneChoice(
        (
            Choice("a:noFill"),
            Choice("a:solidFill"),
            Choice("a:gradFill"),
            Choice("a:blipFill"),
            Choice("a:pattFill"),
            Choice("a:grpFill"),
        ),
        successors=("a:headers", "a:extLst"),
    )
    anchor = OptionalAttribute("anchor", MSO_VERTICAL_ANCHOR)
    marL = OptionalAttribute("marL", ST_Coordinate32)
    marR = OptionalAttribute("marR", ST_Coordinate32)
    marT = OptionalAttribute("marT", ST_Coordinate32)
    marB = OptionalAttribute("marB", ST_Coordinate32)

    def _new_gradFill(self):
        return CT_GradientFillProperties.new_gradFill()


class CT_TableCol(BaseOxmlElement):
    """
    ``<a:gridCol>`` custom element class
    """

    w = RequiredAttribute("w", ST_Coordinate)


class CT_TableGrid(BaseOxmlElement):
    """
    ``<a:tblGrid>`` custom element class
    """

    gridCol = ZeroOrMore("a:gridCol")

    def add_gridCol(self, width):
        """
        Return a reference to a newly created <a:gridCol> child element
        having its ``w`` attribute set to *width*.
        """
        return self._add_gridCol(w=width)


class CT_TableProperties(BaseOxmlElement):
    """
    ``<a:tblPr>`` custom element class
    """

    bandRow = OptionalAttribute("bandRow", XsdBoolean, default=False)
    bandCol = OptionalAttribute("bandCol", XsdBoolean, default=False)
    firstRow = OptionalAttribute("firstRow", XsdBoolean, default=False)
    firstCol = OptionalAttribute("firstCol", XsdBoolean, default=False)
    lastRow = OptionalAttribute("lastRow", XsdBoolean, default=False)
    lastCol = OptionalAttribute("lastCol", XsdBoolean, default=False)


class CT_TableRow(BaseOxmlElement):
    """
    ``<a:tr>`` custom element class
    """

    tc = ZeroOrMore("a:tc", successors=("a:extLst",))
    h = RequiredAttribute("h", ST_Coordinate)

    def add_tc(self):
        """
        Return a reference to a newly added minimal valid ``<a:tc>`` child
        element.
        """
        return self._add_tc()

    @property
    def row_idx(self):
        """Offset of this row in its table."""
        return self.getparent().tr_lst.index(self)

    def _new_tc(self):
        return CT_TableCell.new()


class TcRange(object):
    """A 2D block of `a:tc` cell elements in a table.

    This object assumes the structure of the underlying table does not change
    during its lifetime. Structural changes in this context would be
    insertion or removal of rows or columns.

    The client is expected to create, use, and then abandon an instance in
    the context of a single user operation that is known to have no
    structural side-effects of this type.
    """

    def __init__(self, tc, other_tc):
        self._tc = tc
        self._other_tc = other_tc

    @classmethod
    def from_merge_origin(cls, tc):
        """Return instance created from merge-origin tc element."""
        other_tc = tc.tbl.tc(
            tc.row_idx + tc.rowSpan - 1,  # ---other_row_idx
            tc.col_idx + tc.gridSpan - 1,  # ---other_col_idx
        )
        return cls(tc, other_tc)

    @lazyproperty
    def contains_merged_cell(self):
        """True if one or more cells in range are part of a merged cell."""
        for tc in self.iter_tcs():
            if tc.gridSpan > 1:
                return True
            if tc.rowSpan > 1:
                return True
            if tc.hMerge:
                return True
            if tc.vMerge:
                return True
        return False

    @lazyproperty
    def dimensions(self):
        """(row_count, col_count) pair describing size of range."""
        _, _, width, height = self._extents
        return height, width

    @lazyproperty
    def in_same_table(self):
        """True if both cells provided to constructor are in same table."""
        if self._tc.tbl is self._other_tc.tbl:
            return True
        return False

    def iter_except_left_col_tcs(self):
        """Generate each `a:tc` element not in leftmost column of range."""
        for tr in self._tbl.tr_lst[self._top : self._bottom]:
            for tc in tr.tc_lst[self._left + 1 : self._right]:
                yield tc

    def iter_except_top_row_tcs(self):
        """Generate each `a:tc` element in non-first rows of range."""
        for tr in self._tbl.tr_lst[self._top + 1 : self._bottom]:
            for tc in tr.tc_lst[self._left : self._right]:
                yield tc

    def iter_left_col_tcs(self):
        """Generate each `a:tc` element in leftmost column of range."""
        col_idx = self._left
        for tr in self._tbl.tr_lst[self._top : self._bottom]:
            yield tr.tc_lst[col_idx]

    def iter_tcs(self):
        """Generate each `a:tc` element in this range.

        Cell elements are generated left-to-right, top-to-bottom.
        """
        return (
            tc
            for tr in self._tbl.tr_lst[self._top : self._bottom]
            for tc in tr.tc_lst[self._left : self._right]
        )

    def iter_top_row_tcs(self):
        """Generate each `a:tc` element in topmost row of range."""
        tr = self._tbl.tr_lst[self._top]
        for tc in tr.tc_lst[self._left : self._right]:
            yield tc

    def move_content_to_origin(self):
        """Move all paragraphs in range to origin cell."""
        tcs = list(self.iter_tcs())
        origin_tc = tcs[0]
        for spanned_tc in tcs[1:]:
            origin_tc.append_ps_from(spanned_tc)

    @lazyproperty
    def _bottom(self):
        """Index of row following last row of range"""
        _, top, _, height = self._extents
        return top + height

    @lazyproperty
    def _extents(self):
        """A (left, top, width, height) tuple describing range extents.

        Note this is normalized to accommodate the various orderings of the
        corner cells provided on construction, which may be in any of four
        configurations such as (top-left, bottom-right),
        (bottom-left, top-right), etc.
        """

        def start_and_size(idx, other_idx):
            """Return beginning and length of range based on two indexes."""
            return min(idx, other_idx), abs(idx - other_idx) + 1

        tc, other_tc = self._tc, self._other_tc

        left, width = start_and_size(tc.col_idx, other_tc.col_idx)
        top, height = start_and_size(tc.row_idx, other_tc.row_idx)

        return left, top, width, height

    @lazyproperty
    def _left(self):
        """Index of leftmost column in range"""
        left, _, _, _ = self._extents
        return left

    @lazyproperty
    def _right(self):
        """Index of column following the last column in range"""
        left, _, width, _ = self._extents
        return left + width

    @lazyproperty
    def _tbl(self):
        """`a:tbl` element containing this cell range"""
        return self._tc.tbl

    @lazyproperty
    def _top(self):
        """Index of topmost row in range"""
        _, top, _, _ = self._extents
        return top
